#!/usr/bin/env python2
# coding: utf-8

import os
import pkgutil


def submodules(root_module):

    mod_path = root_module.__file__

    fn = os.path.basename(mod_path)
    pathname = os.path.dirname(mod_path)

    if fn not in ("__init__.py", "__init__.pyc"):
        return None

    rst = {}

    for imp, name, _ in pkgutil.iter_modules([pathname]):
        loader = imp.find_module(name)
        mod = loader.load_module(root_module.__name__ + "." + name)
        rst[name] = mod

    return rst


def submodule_tree(root_module):

    rst = submodules(root_module)
    if rst is None:
        return None

    for name, mod in rst.items():
        children = submodule_tree(mod)
        rst[name] = {"module": mod, "children": children}

    return rst


def submodule_leaf_tree(root_module):

    rst = submodules(root_module)
    if rst is None:
        return None

    for name, mod in rst.items():
        children = submodule_leaf_tree(mod)
        if children is None:
            rst[name] = mod
        else:
            rst[name] = children

    return rst



from . import (
    gtidset,
)

from .mysqlutil import (
    ConnectionTypeError,
    InvalidLength,

    make_delete_sql,
    make_index_scan_sql,
    make_insert_sql,
    make_range_mysqldump_cmd,
    make_select_sql,
    make_sql_range_conditions,
    make_update_sql,
    scan_index,
)

from privilege import (
    privileges
)

__all__ = [
    "ConnectionTypeError",
    "InvalidLength",

    "gtidset",
    "make_delete_sql",
    "make_index_scan_sql",
    "make_insert_sql",
    "make_range_mysqldump_cmd",
    "make_select_sql",
    "make_sql_range_conditions",
    "make_update_sql",
    "privileges",
    "scan_index",
]



import json
import logging
import unittest

from pykit import mysqlutil
from pykit import ututil

dd = ututil.dd

logger = logging.getLogger(__name__)


class TestGtidset(unittest.TestCase):

    maxDiff = 1024

    def test_load_dump(self):

        gtidset_str_a = '\n'.join([
            '103a96e9-a030-11e6-b082-a0369fabbda4:1-20:25,',
            '3f96fb31-6a18-11e6-bd74-a0369fabbdb8:1-30,',
            '630b605d-9ec7-11e6-aae9-a0369fabbdb8:1-90,',
            '83e2adaf-9c37-11e6-9d6e-a0369fb6f7b4:1-70',
        ])

        gtidset_str_b = '\n'.join([
            '103a96e9-a030-11e6-b082-a0369fabbda4:1-20:26,',
            # '3f96fb31-6a18-11e6-bd74-a0369fabbdb8:1-30,',
            '630b605d-9ec7-11e6-aae9-a0369fabbdb8:5-90,',
            '83e2adaf-9c37-11e6-9d6e-a0369fb6f7b4:1-60,',
            'f4ebb415-a801-11e6-9079-a0369fb4eb00:1-80',
        ])

        expected = {
            '103a96e9-a030-11e6-b082-a0369fabbda4': [[1, 20], [25, 25]],
            '3f96fb31-6a18-11e6-bd74-a0369fabbdb8': [[1, 30]],
            '630b605d-9ec7-11e6-aae9-a0369fabbdb8': [[1, 90]],
            '83e2adaf-9c37-11e6-9d6e-a0369fb6f7b4': [[1, 70]],
        }

        dd()
        dd(gtidset_str_a)

        dd('test load')

        gs_a = mysqlutil.gtidset.load(gtidset_str_a)
        dd('gs_a:', gs_a)
        self.assertEqual(expected, gs_a)

        dd('test dump')

        gs_str_a = mysqlutil.gtidset.dump(gs_a)
        dd('gs_str_a:', gs_str_a)
        self.assertEqual(gtidset_str_a, gs_str_a)
        self.assertEqual(gtidset_str_a.replace('\n', ''),
                         mysqlutil.gtidset.dump(gs_a, line_break=''))

        dd('test compare gtidset')

        expected_cmp_rst = {
            "onlyleft": {
                'length': 45,
                'gtidset': {
                    "103a96e9-a030-11e6-b082-a0369fabbda4": [[25, 25]],
                    "3f96fb31-6a18-11e6-bd74-a0369fabbdb8": [[1, 30]],
                    "630b605d-9ec7-11e6-aae9-a0369fabbdb8": [[1, 4]],
                    "83e2adaf-9c37-11e6-9d6e-a0369fb6f7b4": [[61, 70]]
                },
            },
            "onlyright": {
                'length': 81,
                'gtidset': {
                    "103a96e9-a030-11e6-b082-a0369fabbda4": [[26, 26]],
                    "f4ebb415-a801-11e6-9079-a0369fb4eb00": [[1, 80]]
                },
            },
        }

        gs_b = mysqlutil.gtidset.load(gtidset_str_b)

        cmp_rst = mysqlutil.gtidset.compare(gs_a, gs_b)
        dd(json.dumps(expected_cmp_rst, indent=2))
        dd(json.dumps(cmp_rst, indent=2))

        self.assertEqual(expected_cmp_rst, cmp_rst)



from .client import (
    HttpError,
    LineTooLongError,
    ChunkedSizeError,
    NotConnectedError,
    ResponseNotReadyError,
    HeadersError,
    BadStatusLineError,
    Client,
)

__all__ = [
    'HttpError',
    'LineTooLongError',
    'ChunkedSizeError',
    'NotConnectedError',
    'ResponseNotReadyError',
    'HeadersError',
    'BadStatusLineError',
    'Client',
]



import logging
import re
import socket
import struct
import sys
import types

import netifaces
import yaml

PUB = 'PUB'
INN = 'INN'

LOCALHOST = '127.0.0.1'

logger = logging.getLogger(__name__)

_intra_patterns = (
    '172\.(1[6-9]|2[0-9]|3[0-1]).*',
    '10\..*',
    '192\.168\..*',
)


class NetworkError(Exception):
    pass


class IPUnreachable(NetworkError):
    pass


class InvalidIP4(Exception):
    pass


class InvalidIP4Number(Exception):
    pass


def is_ip4(ip):

    if type(ip) not in types.StringTypes:
        return False

    ip = ip.split('.')

    for s in ip:

        if not s.isdigit():
            return False

        i = int(s)
        if i < 0 or i > 255:
            return False

    return len(ip) == 4


def ip_class(ip):

    if ip.startswith('127.0.0.'):
        return INN

    for ptn in _intra_patterns:

        if re.match(ptn, ip):
            return INN

    else:
        return PUB


def is_ip4_loopback(ip):

    return is_ip4(ip) and ip.startswith('127.')


def ips_prefer(ips, preference):

    eips = choose_pub(ips)
    iips = choose_inn(ips)

    if preference == PUB:
        return eips + iips
    else:
        return iips + eips


def is_pub(ip):
    return ip_class(ip) == PUB


def is_inn(ip):
    return ip_class(ip) == INN


def choose_pub(ips):
    return [x for x in ips if ip_class(x) == PUB]


def choose_inn(ips):
    return [x for x in ips if ip_class(x) == INN]


def get_host_ip4(iface_prefix=None, exclude_prefix=None):

    if iface_prefix is None:
        iface_prefix = ['']

    if type(iface_prefix) in types.StringTypes:
        iface_prefix = [iface_prefix]

    if exclude_prefix is not None:
        if type(exclude_prefix) in types.StringTypes:
            exclude_prefix = [exclude_prefix]

    ips = []

    for ifacename in netifaces.interfaces():

        matched = False

        for t in iface_prefix:
            if ifacename.startswith(t):
                matched = True
                break

        if exclude_prefix is not None:
            for ex in exclude_prefix:
                if ifacename.startswith(ex):
                    matched = False
                    break

        if not matched:
            continue

        addrs = netifaces.ifaddresses(ifacename)

        if netifaces.AF_INET in addrs and netifaces.AF_LINK in addrs:

            for addr in addrs[netifaces.AF_INET]:

                ip = str(addr['addr'])

                if not is_ip4_loopback(ip):
                    ips.append(ip)

    return ips


def choose_by_idc(dest_idc, local_idc, ips):

    if dest_idc == local_idc:
        pref_ips = ips_prefer(ips, INN)
    else:
        pref_ips = ips_prefer(ips, PUB)

    return pref_ips


def get_host_devices(iface_prefix=''):

    rst = {}

    for ifacename in netifaces.interfaces():

        if not ifacename.startswith(iface_prefix):
            continue

        addrs = netifaces.ifaddresses(ifacename)

        if netifaces.AF_INET in addrs and netifaces.AF_LINK in addrs:

            ips = [addr['addr'] for addr in addrs[netifaces.AF_INET]]

            for ip in ips:
                if is_ip4_loopback(ip):
                    break
            else:
                rst[ifacename] = {'INET': addrs[netifaces.AF_INET],
                                  'LINK': addrs[netifaces.AF_LINK]}

    return rst


def parse_ip_regex_str(ip_regexs_str):

    ip_regexs_str = ip_regexs_str.strip()

    regs = ip_regexs_str.split(',')
    rst = []
    for r in regs:
        # do not choose ip if it matches this regex
        if r.startswith('-'):
            r = (r[1:], False)
        else:
            r = (r, True)

        if r[0] == '':
            raise ValueError('invalid regular expression: ' + repr(r))

        if r[1]:
            r = r[0]

        rst.append(r)

    return rst


def choose_by_regex(ips, ip_regexs):

    rst = []

    for ip in ips:

        all_negative = True
        for ip_regex in ip_regexs:

            # choose matched:
            #     '127[.]'
            #     ('127[.]', True)
            # choose unmatched:
            #     ('127[.], False)

            if type(ip_regex) in (type(()), type([])):
                ip_regex, to_choose = ip_regex
            else:
                ip_regex, to_choose = ip_regex, True

            all_negative = all_negative and not to_choose

            # when to choose it:
            #     match one of positive regex.
            #     and match none of negative regex.

            if re.match(ip_regex, ip):
                if to_choose:
                    rst.append(ip)

                break
        else:
            # if all regexs are for excluding ip, then choose it
            if all_negative:
                rst.append(ip)

    return rst


def ip_to_num(ip_str):

    if not is_ip4(ip_str):
        raise InvalidIP4('IP is invalid: {s}'.format(s=ip_str))

    return struct.unpack('>L', socket.inet_aton(ip_str))[0]


def num_to_ip(ip_num):

    if isinstance(ip_num, bool) or not isinstance(ip_num, (int, long)):
        raise InvalidIP4Number('The type of IP4 number should be int or long :{t}'.format(t=type(ip_num)))
    if ip_num > 0xffffffff or ip_num < 0:
        raise InvalidIP4Number('IP4 number should be between 0 and 0xffffffff :{s}'.format(s=ip_num))

    return socket.inet_ntoa(struct.pack('>L', ip_num))


if __name__ == "__main__":

    args = sys.argv[1:]

    if args[0] == 'ip':
        print yaml.dump(get_host_ip4(), default_flow_style=False)

    elif args[0] == 'device':
        print yaml.dump(get_host_devices(), default_flow_style=False)

    else:
        raise ValueError('invalid command line arguments', args)



import os
import time
import unittest
import uuid

from kazoo import security
from kazoo.client import KazooClient
from kazoo.exceptions import ConnectionClosedError
from kazoo.exceptions import NoNodeError

from pykit import config
from pykit import net
from pykit import threadutil
from pykit import utdocker
from pykit import ututil
from pykit import zkutil

dd = ututil.dd

zk_tag = 'daocloud.io/zookeeper:3.4.10'
zk_name = 'zk_test'


class TestZKutil(unittest.TestCase):

    def test_lock_id(self):

        k = zkutil.lock_id('a')
        dd(k)
        elts = k.split('-')

        self.assertEqual(4, len(elts))

        self.assertEqual('a', elts[0])
        self.assertTrue(net.is_ip4(elts[1]))
        self.assertEqual(os.getpid(), int(elts[2]))

    def test_lock_id_default(self):

        expected = '%012x' % uuid.getnode()

        k = zkutil.lock_id()
        dd(config)
        dd(k)
        self.assertEqual(expected, k.split('-')[0])

        k = zkutil.lock_id(node_id=None)
        dd(k)
        self.assertEqual(expected, k.split('-')[0])

        config.zk_node_id = 'a'
        k = zkutil.lock_id(node_id=None)
        dd(k)
        self.assertEqual('a', k.split('-')[0])

    def test_parse_lock_id(self):

        cases = (
                ('', ('', None, None)),
                ('-', ('', '', None)),
                ('--', ('', '', None)),
                ('1-', ('1', '', None)),
                ('1-a', ('1', 'a', None)),
                ('1-a-x', ('1', 'a', None)),
                ('1-a-1', ('1', 'a', 1)),
        )

        for inp, expected in cases:

            rst = zkutil.parse_lock_id(inp)

            self.assertEqual(set(['node_id', 'ip', 'process_id', 'uuid', 'txid']),
                             set(rst.keys()))

            self.assertEqual(expected,
                             (rst['node_id'], rst['ip'], rst['process_id']))

    def test_parse_lock_id_with_txid(self):

        rst = zkutil.parse_lock_id('txid:123-a-b-x')
        self.assertEqual('txid:123', rst['node_id'])
        self.assertEqual('123', rst['txid'])

    def test_make_digest(self):

        cases = (
                ('aa:', 'E/ZoYMT80fFT7vhICWyvMdWNt7o='),
                ('username:password', '+Ir5sN1lGJEEs8xBZhZXKvjLJ7c=')
        )

        for inp, expected in cases:
            rst = zkutil.make_digest(inp)
            self.assertEqual(expected, rst)

    def test_make_acl_entry(self):

        username = 'zookeeper_user'
        password = 'MT80fFT7vh'
        perm_cases = [
            '',
            'rw',
            'cdrwa',
            [],
            ['c'],
            ['c', 'r', 'r'],
            (),
            ('c',),
            ('c', 'r', 'w'),
            iter('cdrwa'),
        ]

        for inp in perm_cases:

            if isinstance(inp, type(iter(''))):
                perm = 'cdrwa'
            else:
                perm = ''.join(inp)

            dd('inp=', inp)
            dd('perm=', perm)
            rst = zkutil.make_acl_entry(username, password, inp)

            rst_splited = rst.split(':')
            self.assertEqual(4, len(rst_splited))
            self.assertEqual(('digest', username, perm),
                             (rst_splited[0], rst_splited[1], rst_splited[3]))

        cases = (
                ((username, password, ''),
                 'digest:zookeeper_user:Ds8aM7UNwfAlTN3IRdkBoCno9FM=:'),
                (('aabb', 'abc123', 'cdrwa'),
                 'digest:aabb:t9MeAsoEPfdQEQjqbWtw8EHD9T0=:cdrwa')
        )

        for inp, expected in cases:
            rst = zkutil.make_acl_entry(inp[0], inp[1], inp[2])
            self.assertEqual(expected, rst)

        invalid_perm_cases = (
            'abc',
            ['cde'],
            ['a', 'v'],
            ('rw',),
            ('a', 'b', 'c'),
        )

        for inp in invalid_perm_cases:

            with self.assertRaises(zkutil.PermTypeError):
                zkutil.make_acl_entry(username, password, inp)

    def test_permission_convert(self):

        perm_cases = (
            ('',              [], ''),
            ('rw',            ['read', 'write'], 'rw'),
            ('cdrwa',         ['create', 'delete',
                               'read', 'write', 'admin'], 'cdrwa'),
            ([],              [], ''),
            (['c'],           ['create'], 'c'),
            (['c', 'r', 'r'], ['create', 'read', 'read'], 'crr'),
            ((),              [], ''),
            (('c',),          ['create'], 'c'),
            (('c', 'r', 'w'), ['create', 'read', 'write'], 'crw'),
            (iter('cdrwa'),   ['create', 'delete',
                               'read', 'write', 'admin'], 'cdrwa'),
        )

        for inp, lng, short in perm_cases:

            rst = zkutil.perm_to_long(inp)
            self.assertEqual(lng, rst)

            rst = zkutil.perm_to_short(lng)
            self.assertEqual(short, rst)

        # invalid short format

        invalid_short = (
            'abc',
            ['cde'],
            ['a', 'v'],
            ('rw',),
            ('a', 'b', 'c'),
        )

        for inp in invalid_short:
            self.assertRaises(zkutil.PermTypeError, zkutil.perm_to_long, inp)

        # invalid long format

        invalid_long = (
            'abc',
            ['foo'],
        )

        for inp in invalid_long:
            self.assertRaises(zkutil.PermTypeError, zkutil.perm_to_short, inp)

    def test_permission_convert_case(self):

        self.assertEqual(['CREATE', 'DELETE', 'READ', 'WRITE', 'ADMIN'],
                         zkutil.perm_to_long('cdrwa', lower=False))

        self.assertEqual(['CREATE', 'DELETE', 'READ', 'WRITE', 'ADMIN'],
                         zkutil.perm_to_long('CDRWA', lower=False))

        self.assertEqual(['create', 'delete', 'read', 'write', 'admin'],
                         zkutil.perm_to_long('CDRWA'))

        self.assertEqual('CDRWA',
                         zkutil.perm_to_short(['create', 'delete', 'read', 'write', 'admin'],
                                              lower=False))

        self.assertEqual('CDRWA',
                         zkutil.perm_to_short(['CREATE', 'DELETE', 'READ', 'WRITE', 'ADMIN'],
                                              lower=False))
        self.assertEqual('cdrwa',
                         zkutil.perm_to_short(['CREATE', 'DELETE', 'READ', 'WRITE', 'ADMIN']))

    def test_make_kazoo_digest_acl(self):

        inp = (('foo', 'bar', 'cd'),
               ('xp', '123', 'cdrwa'))

        dd(inp)

        rst = zkutil.make_kazoo_digest_acl(inp)
        dd(rst)

        self.assertEqual(2, len(rst))

        ac = rst[0]
        self.assertEqual('digest', ac.id.scheme)
        self.assertEqual('foo', ac.id.id.split(':')[0])
        self.assertEqual(set(['CREATE', 'DELETE']), set(ac.acl_list))

        ac = rst[1]
        self.assertEqual('digest', ac.id.scheme)
        self.assertEqual('xp', ac.id.id.split(':')[0])
        self.assertEqual(set(['ALL']), set(ac.acl_list))

        self.assertIsNone(zkutil.make_kazoo_digest_acl(None))

    def test_parse_kazoo_acl(self):

        inp = (security.make_acl('world', 'anyone', all=True),
               security.make_digest_acl('foo', 'bar', create=True, read=True),
               security.make_digest_acl('xp', '123', all=True),
               )
        expected = (('world', 'anyone', 'cdrwa'),
                    ('digest', 'foo', 'rc'),
                    ('digest', 'xp', 'cdrwa'))

        dd(inp)

        rst = zkutil.parse_kazoo_acl(inp)
        dd(rst)

        self.assertEqual(expected, tuple(rst))

    def test_is_backward_locking(self):

        cases = (
            ([],         'a', False, None),
            (['a'],      'a', False, AssertionError),
            (['a', 'c'], 'a', True,  AssertionError),
            (['a', 'c'], 'c', True,  AssertionError),
            (['a', 'c'], '',  True,  None),
            (['a', 'c'], 'b', True,  None),
            (['a', 'c'], 'd', False, None),
        )

        for locked, key, expected, err in cases:

            if err is None:
                rst = zkutil.is_backward_locking(locked, key)
                self.assertEqual(expected, rst)
            else:
                self.assertRaises(err, zkutil.is_backward_locking, locked, key)


class TestZKinit(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        utdocker.pull_image(zk_tag)

    def setUp(self):

        utdocker.create_network()

        utdocker.start_container(
            zk_name,
            zk_tag,
            port_bindings={2181: 2181}
        )

        dd('start zk-test in docker')

    def tearDown(self):

        utdocker.stop_container(zk_name)
        utdocker.remove_container(zk_name)

        dd('remove_container: ' + zk_name)

    def test_init_hierarchy(self):

        auth = ('digest', 'aa', 'pw_aa')
        hosts = '127.0.0.1:2181'
        users = {'aa': 'pw_aa', 'bb': 'pw_bb', 'cc': 'pw_cc'}
        hierarchy = {
            'node1':
            {
                '__val__': 'node1_val',
                '__acl__': {'aa': 'cdrwa', 'bb': 'rw'},
                'node11':
                {
                    '__val__': 'node11_val',
                    '__acl__': {'aa': 'cdrwa', 'cc': 'r'},
                },
                'node12':
                {
                    '__val__': 'node12_val',
                    'node121': {'__val__': 'node121_val'}
                },
                'node13':
                {
                    '__acl__': {'aa': 'cdrwa'}
                }
            },
            'node2':
            {
                '__val__': 'node2_val',
                'node21': {'__val__': 'node21_val'},
                'node22': {'__acl__': {'aa': 'rwa'}}
            },
            'node3':
            {
                '__acl__': {'aa': 'carw', 'cc': 'r'},
                'node31': {'node311': {'node3111': {}, 'node3112': {}}}
            }
        }

        zkutil.init_hierarchy(hosts, hierarchy, users, auth)

        zkcli = KazooClient(hosts)
        zkcli.start()
        zkcli.add_auth('digest', 'aa:pw_aa')

        expected_nodes = (
            ('/node1', '"node1_val"', [('digest', 'aa', 'cdrwa'),
                                       ('digest', 'bb', 'rw')], set(['node11', 'node12', 'node13'])),
            ('/node1/node11', '"node11_val"',
             [('digest', 'aa', 'cdrwa'), ('digest', 'cc', 'r')], set([])),
            ('/node1/node12', '"node12_val"',
             [('digest', 'aa', 'cdrwa'), ('digest', 'bb', 'rw')], set(['node121'])),
            ('/node1/node12/node121', '"node121_val"',
             [('digest', 'aa', 'cdrwa'), ('digest', 'bb', 'rw')], set([])),
            ('/node1/node13', '{}', [('digest', 'aa', 'cdrwa')], set([])),

            ('/node2', '"node2_val"',
             [('world', 'anyone', 'cdrwa')], set(['node21', 'node22'])),
            ('/node2/node21', '"node21_val"',
             [('world', 'anyone', 'cdrwa')],  set([])),
            ('/node2/node22', '{}', [('digest', 'aa', 'rwa')],  set([])),

            ('/node3', '{}', [('digest', 'aa', 'rwca'),
                              ('digest', 'cc', 'r')], set(['node31'])),
            ('/node3/node31', '{}', [('digest', 'aa', 'rwca'),
                                     ('digest', 'cc', 'r')],  set(['node311'])),
            ('/node3/node31/node311', '{}',
             [('digest', 'aa', 'rwca'), ('digest', 'cc', 'r')], set(['node3111', 'node3112'])),
            ('/node3/node31/node311/node3111', '{}',
             [('digest', 'aa', 'rwca'), ('digest', 'cc', 'r')], set([])),
            ('/node3/node31/node311/node3112', '{}',
             [('digest', 'aa', 'rwca'), ('digest', 'cc', 'r')], set([])),
        )

        for node, val, acl, children in expected_nodes:

            actual_acl = zkutil.parse_kazoo_acl(zkcli.get_acls(node)[0])
            self.assertEqual(val, zkcli.get(node)[0])
            self.assertEqual(acl, actual_acl)
            self.assertEqual(children, set(zkcli.get_children(node)))

        zkcli.stop()


class TestWait(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        utdocker.pull_image(zk_tag)

    def setUp(self):

        utdocker.create_network()
        utdocker.start_container(
            zk_name,
            zk_tag,
            env={
                "ZOO_MY_ID": 1,
                "ZOO_SERVERS": "server.1=0.0.0.0:2888:3888",
            },
            port_bindings={2181: 2181}
        )

        self.zk = KazooClient('127.0.0.1:2181')
        self.zk.start()

        dd('start zk-test in docker')

    def tearDown(self):

        self.zk.stop()
        utdocker.remove_container(zk_name)

    def test_wait_absent(self):

        for wait_time in (
                -1,
                0.0,
                0.1,
                1,
        ):

            dd('no node wait:', wait_time)

            with ututil.Timer() as t:
                zkutil.wait_absent(self.zk, 'a', wait_time)
                self.assertAlmostEqual(0, t.spent(), delta=0.2)

    def test_wait_absent_no_timeout(self):

        def _del():
            time.sleep(1)
            self.zk.delete('a')

        for kwargs in (
            {},
                {'timeout': None},
        ):

            self.zk.create('a')
            th = threadutil.start_daemon(target=_del)

            with ututil.Timer() as t:
                zkutil.wait_absent(self.zk, 'a', **kwargs)
                self.assertAlmostEqual(1, t.spent(), delta=0.1)

            th.join()

    def test_wait_absent_timeout(self):

        self.zk.create('a')

        for wait_time in (
                -1,
                0.0,
                0.1,
                1,
        ):
            dd('node present wait:', wait_time)
            expected = max([0, wait_time])

            with ututil.Timer() as t:
                self.assertRaises(zkutil.ZKWaitTimeout,
                                  zkutil.wait_absent,
                                  self.zk, 'a', timeout=wait_time)
                self.assertAlmostEqual(expected, t.spent(), delta=0.2)

        self.zk.delete('a')

    def test_wait_absent_delete_node(self):

        delete_after = 0.2

        for wait_time in (
                0.5,
                1,
        ):
            dd('node present wait:', wait_time)

            self.zk.create('a')

            def _del():
                time.sleep(delete_after)
                self.zk.delete('a')

            th = threadutil.start_daemon(target=_del)
            with ututil.Timer() as t:
                zkutil.wait_absent(self.zk, 'a', wait_time)
                self.assertAlmostEqual(delete_after, t.spent(), delta=0.1)

            th.join()

    def test_wait_absent_change_node(self):

        self.zk.create('a')

        change_after = 0.2

        for wait_time in (
                0.5,
                1,
        ):
            dd('node present wait:', wait_time)
            expected = max([0, wait_time])

            def _change():
                time.sleep(change_after)
                self.zk.set('a', 'bbb')

            th = threadutil.start_daemon(target=_change)
            with ututil.Timer() as t:
                self.assertRaises(zkutil.ZKWaitTimeout,
                                  zkutil.wait_absent,
                                  self.zk, 'a', timeout=wait_time)
                self.assertAlmostEqual(expected, t.spent(), delta=0.1)

            th.join()

        self.zk.delete('a')

    def test_wait_absent_connection_lost(self):

        self.zk.create('a')

        def _close():
            time.sleep(.3)
            self.zk.stop()

        th = threadutil.start_daemon(target=_close)

        with ututil.Timer() as t:
            self.assertRaises(ConnectionClosedError,
                              zkutil.wait_absent,
                              self.zk, 'a')
            self.assertAlmostEqual(.3, t.spent(), delta=0.1)

        th.join()

    def test_get_next_no_version(self):

        cases = (
            -1,
            0.0,
            0.1,
            1,
        )

        for timeout in cases:

            self.zk.create('a', 'a-val')

            with ututil.Timer() as t:
                zkutil.get_next(self.zk, 'a', timeout=timeout, version=-1)
                self.assertAlmostEqual(0, t.spent(), delta=0.2)

            with ututil.Timer() as t:
                zkutil.get_next(self.zk, 'a', timeout=timeout)
                self.assertAlmostEqual(0, t.spent(), delta=0.2)

            self.zk.delete('a')

    def test_get_next_timeout(self):

        cases = (
            -1,
            0.0,
            0.2,
            1,
        )

        for timeout in cases:

            expected = max([timeout, 0])
            self.zk.create('a', 'a-val')

            with ututil.Timer() as t:
                self.assertRaises(zkutil.ZKWaitTimeout,
                                  zkutil.get_next,
                                  self.zk, 'a', timeout=timeout, version=0)
                self.assertAlmostEqual(expected, t.spent(), delta=0.2)

            self.zk.delete('a')

    def test_get_next_changed(self):

        cases = (
            0.4,
            1,
        )

        def _set_a():
            self.zk.set('a', 'changed')

        for timeout in cases:

            self.zk.create('a', 'a-val')
            th = threadutil.start_daemon(target=_set_a, after=0.3)

            with ututil.Timer() as t:
                val, zstat = zkutil.get_next(self.zk, 'a', timeout=timeout, version=0)
                self.assertAlmostEqual(0.3, t.spent(), delta=0.2)
                self.assertEqual('changed', val)
                self.assertEqual(1, zstat.version)

            th.join()
            self.zk.delete('a')

    def test_get_next_changed_but_unsatisfied(self):

        cases = (
            0.4,
            1,
        )

        def _set_a():
            self.zk.set('a', 'changed')

        for timeout in cases:

            self.zk.create('a', 'a-val')
            th = threadutil.start_daemon(target=_set_a, after=0.3)

            with ututil.Timer() as t:
                self.assertRaises(zkutil.ZKWaitTimeout,
                                  zkutil.get_next,
                                  self.zk, 'a', timeout=timeout, version=5)
                self.assertAlmostEqual(timeout, t.spent(), delta=0.2)

            th.join()
            self.zk.delete('a')

    def test_get_next_deleted(self):

        cases = (
            0.4,
            1,
        )

        def _del_a():
            self.zk.delete('a')

        for timeout in cases:

            self.zk.create('a', 'a-val')
            th = threadutil.start_daemon(target=_del_a, after=0.3)

            with ututil.Timer() as t:
                self.assertRaises(NoNodeError,
                                  zkutil.get_next,
                                  self.zk, 'a', timeout=timeout, version=0)
                self.assertAlmostEqual(0.3, t.spent(), delta=0.2)

            th.join()

    def test_get_next_conn_lost(self):

        self.zk.create('a', 'a-val')
        th = threadutil.start_daemon(target=self.zk.stop, after=0.3)

        with ututil.Timer() as t:
            self.assertRaises(ConnectionClosedError,
                              zkutil.get_next,
                              self.zk, 'a', timeout=1, version=0)
            self.assertAlmostEqual(0.3, t.spent(), delta=0.2)

        th.join()



#!/usr/bin/env python2
# coding: utf-8

import time
import unittest

from pykit import etcd
from pykit import utdocker
from pykit import utfjson
from pykit import ututil

dd = ututil.dd

HOSTS = (
    ("192.168.52.30", 3379),
    ("192.168.52.31", 3379),
    ("192.168.52.32", 3379),
    ("192.168.52.33", 3379),
    ("192.168.52.34", 3379),
)

ETCD_NAMES = ('etcd_t0', 'etcd_t1', 'etcd_t2', 'etcd_t3', 'etcd_t4',)
ETCD_NODES = ('node_0', 'node_1', 'node_2', 'node_3', 'node_4',)

etcd_test_tag = 'quay.io/coreos/etcd:v2.3.7'


class TestEtcdKeysResult(unittest.TestCase):

    def test_get_subtree_1_level(self):

        response = {"node": {
            'key': "/test",
            'value': "hello",
            'expiration': None,
            'ttl': None,
            'modifiedIndex': 5,
            'createdIndex': 1,
            'newKey': False,
            'dir': False,
        }}
        result = etcd.EtcdKeysResult(**response)
        for k, v in response['node'].items():
            self.assertEqual(v, getattr(result, k))

        self.assertListEqual([result], list(
            result.get_subtree(leaves_only=False)))
        self.assertListEqual([result], list(
            result.get_subtree(leaves_only=True)))
        self.assertListEqual([result], list(result.leaves))

    def test_get_subtree_2_level(self):

        leaf0 = {
            'key': "/test/leaf0",
            'value': "hello1",
            'expiration': None,
            'ttl': None,
            'modifiedIndex': 5,
            'createdIndex': 1,
            'newKey': False,
            'dir': False,
        }
        leaf1 = {
            'key': "/test/leaf1",
            'value': "hello2",
            'expiration': None,
            'ttl': None,
            'modifiedIndex': 6,
            'createdIndex': 2,
            'newKey': False,
            'dir': False,
        }
        testnode = {"node": {
            'key': "/test/",
            'expiration': None,
            'ttl': None,
            'modifiedIndex': 6,
            'createdIndex': 2,
            'newKey': False,
            'dir': True,
            'nodes': [leaf0, leaf1]
        }}
        result = etcd.EtcdKeysResult(**testnode)
        for k, v in testnode['node'].items():
            if not hasattr(result, k):
                continue

            self.assertEqual(v, getattr(result, k))

        self.assertListEqual(result._children, testnode['node']['nodes'])

        subtree = list(result.get_subtree(leaves_only=True))
        self.assertEqual("/test/leaf0", subtree[0].key)
        self.assertEqual("/test/leaf1", subtree[1].key)
        self.assertEqual(len(subtree), 2)
        self.assertListEqual(subtree, list(result.leaves))

        subtree = list(result.get_subtree(leaves_only=False))
        self.assertEqual("/test/", subtree[0].key)
        self.assertEqual("/test/leaf0", subtree[1].key)
        self.assertEqual("/test/leaf1", subtree[2].key)
        self.assertEqual(len(subtree), 3)

    def test_get_subtree_3_level(self):

        leaf0 = {
            'key': "/test/mid0/leaf0",
            'value': "hello1",
        }
        leaf1 = {
            'key': "/test/mid0/leaf1",
            'value': "hello2",
        }
        leaf2 = {
            'key': "/test/mid1/leaf2",
            'value': "hello1",
        }
        leaf3 = {
            'key': "/test/mid1/leaf3",
            'value': "hello2",
        }
        mid0 = {
            'key': "/test/mid0/",
            'dir': True,
            'nodes': [leaf0, leaf1]
        }
        mid1 = {
            'key': "/test/mid1/",
            'dir': True,
            'nodes': [leaf2, leaf3]
        }
        testnode = {"node": {
            'key': "/test/",
            'dir': True,
            'nodes': [mid0, mid1]
        }}

        result = etcd.EtcdKeysResult(**testnode)
        for k, v in testnode['node'].items():
            if not hasattr(result, k):
                continue

            self.assertEqual(v, getattr(result, k))

        self.assertListEqual(result._children, testnode['node']['nodes'])

        subtree = list(result.get_subtree(leaves_only=True))
        self.assertEqual("/test/mid0/leaf0", subtree[0].key)
        self.assertEqual("/test/mid0/leaf1", subtree[1].key)
        self.assertEqual("/test/mid1/leaf2", subtree[2].key)
        self.assertEqual("/test/mid1/leaf3", subtree[3].key)
        self.assertEqual(len(subtree), 4)
        self.assertListEqual(subtree, list(result.leaves))

        subtree = list(result.get_subtree(leaves_only=False))
        self.assertEqual("/test/", subtree[0].key)
        self.assertEqual("/test/mid0/", subtree[1].key)
        self.assertEqual("/test/mid0/leaf0", subtree[2].key)
        self.assertEqual("/test/mid0/leaf1", subtree[3].key)
        self.assertEqual("/test/mid1/", subtree[4].key)
        self.assertEqual("/test/mid1/leaf2", subtree[5].key)
        self.assertEqual("/test/mid1/leaf3", subtree[6].key)
        self.assertEqual(len(subtree), 7)

    def test_eq_ne(self):

        cases = (
            ({'node': {'key': 'key1'}},
             {'node': {'key': 'key2'}},
             False,
             ),

            ({'node': {'key': 'key1'}},
             {'node': {'key': 'key1'}},
             True,
             ),

            ({'node': {'key': 'key1', 'value': 'val1'}},
             {'node': {'key': 'key1', 'value': 'val2'}},
             False,
             ),
        )

        for node1, node2, expected_res in cases:
            res1 = etcd.EtcdKeysResult(**node1)
            res2 = etcd.EtcdKeysResult(**node2)
            self.assertEqual(res1 == res2, expected_res)


class TestException(unittest.TestCase):

    def test_errorcode_exception(self):

        cases = (
            ({'errorCode': 100, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeKeyNotFound, 'foo : bar'),

            ({'errorCode': 101, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeTestFailed, 'foo : bar'),

            ({'errorCode': 102, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeNotFile, 'foo : bar'),

            ({'errorCode': 103, 'message': 'foo', 'cause': 'bar'},
             etcd.EtcdException, 'foo : bar'),

            ({'errorCode': 104, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeNotDir, 'foo : bar'),

            ({'errorCode': 105, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeNodeExist, 'foo : bar'),

            ({'errorCode': 106, 'message': 'foo', 'cause': 'bar'},
             etcd.EtcdKeyError, 'foo : bar'),

            ({'errorCode': 107, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeRootROnly, 'foo : bar'),

            ({'errorCode': 108, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeDirNotEmpty, 'foo : bar'),

            ({'errorCode': 110, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeInscientPermissions, 'foo : bar'),

            ({'errorCode': 200, 'message': 'foo', 'cause': 'bar'},
             etcd.EtcdValueError, 'foo : bar'),

            ({'errorCode': 201, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodePrevValueRequired, 'foo : bar'),

            ({'errorCode': 202, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeTTLNaN, 'foo : bar'),

            ({'errorCode': 203, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeIndexNaN, 'foo : bar'),

            ({'errorCode': 209, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeInvalidField, 'foo : bar'),

            ({'errorCode': 210, 'message': 'foo', 'cause': 'bar'},
             etcd.EcodeInvalidForm, 'foo : bar'),

            ({'errorCode': 300, 'message': 'foo', 'cause': 'bar'},
             etcd.EtcdInternalError, 'foo : bar'),

            ({'errorCode': 301, 'message': 'foo', 'cause': 'bar'},
             etcd.EtcdInternalError, 'foo : bar'),

            ({'errorCode': 400, 'message': 'foo', 'cause': 'bar'},
             etcd.EtcdWatchError, 'foo : bar'),

            ({'errorCode': 401, 'message': 'foo', 'cause': 'bar'},
             etcd.EtcdWatchError, 'foo : bar'),

            ({'errorCode': 500, 'message': 'foo', 'cause': 'bar'},
             etcd.EtcdInternalError, 'foo : bar'),

            ({'errorCode': 501, 'message': 'foo', 'cause': 'bar'},
             etcd.EtcdException, 'Unable to decode server response'),

            ({},
             etcd.EtcdException, 'Unable to decode server response'),
        )

        for res_body, e, expect_msg in cases:
            res = etcd.Response(body=utfjson.dump(res_body))
            self.assertRaisesRegexp(e, expect_msg, etcd.EtcdError.handle, res)


class TestClient(unittest.TestCase):

    def setUp(self):
        _start_etcd_server(HOSTS, ETCD_NAMES, ETCD_NODES)
        self._clear_all()

    def _clear_all(self):

        while True:
            try:
                c = etcd.Client(host=HOSTS)
                res = c.get('', recursive=True)
                for n in res._children:
                    c.delete(n['key'], dir='dir' in n and n['dir'],
                             recursive=True)
                break
            except Exception as e:
                dd(repr(e))
                time.sleep(1)

    def test_machine_cache(self):

        machine = list(HOSTS)
        machine.pop(0)
        machine = ['http://%s:%d' % (ip, port) for ip, port in machine]
        cases = (
            (HOSTS, 3379, True, machine),
            ('192.168.52.30', 3379, True, machine),
            (HOSTS, 3379, False, []),
            ('192.168.52.30', 3379, False, []),
        )

        for host, port, allow_reconnect, expected_machines in cases:
            c = etcd.Client(host=host, port=port,
                            allow_reconnect=allow_reconnect)
            self.assertEqual('http://192.168.52.30:3379', c._base_uri)
            self.assertListEqual(expected_machines, c._machines_cache)

    def test_property(self):

        cases = (
            ('192.168.52.30', 3379, 'http', 10, True),
            ('192.168.52.31', 3379, 'http', 20, False),
        )

        for host, port, protocol, timeout, allow_redirect in cases:
            c = etcd.Client(host=host, port=port, protocol=protocol,
                            read_timeout=timeout, allow_redirect=allow_redirect)
            self.assertEqual(host, c.host)
            self.assertEqual(port, c.port)
            self.assertEqual(protocol, c.protocol)
            self.assertEqual(timeout, c.read_timeout)
            self.assertEqual(allow_redirect, c.allow_redirect)
            self.assertEqual('%s://%s:%d' % (protocol, host, port), c.base_uri)

    def test_members(self):

        c = etcd.Client(host=HOSTS)

        mems = c.members
        for h, p in HOSTS:
            succ = False
            for m in mems:
                if 'http://%s:%d' % (h, p) == m['clientURLs'][0]:
                    succ = True
                    break
            self.assertTrue(succ)

    def test_leader(self):

        c = etcd.Client(host=HOSTS)

        leader = c.leader
        self.assertIsNotNone(leader)
        self.assertIn(leader, c.members)

    def test_version(self):

        c = etcd.Client(host=HOSTS)

        ver = c.version
        self.assertIn('etcdserver', ver)
        self.assertIn('etcdcluster', ver)

    def test_nomoremachine_exception(self):

        utdocker.stop_container(*ETCD_NAMES)
        c = etcd.Client(host=HOSTS, read_timeout=1)
        self.assertRaises(etcd.NoMoreMachineError, c.get, '/key')

    def test_etcdreadtimeouterror_exception(self):

        utdocker.stop_container(*ETCD_NAMES)
        c = etcd.Client(host=HOSTS, read_timeout=1)
        self.assertRaises(etcd.EtcdReadTimeoutError, c.api_execute,
                          '/v2/keys/abc', 'GET', timeout=1,
                          raise_read_timeout=True)

    def test_etcdrequestserror_exception(self):

        c = etcd.Client(host=HOSTS)
        self.assertRaises(etcd.EtcdRequestError, c._request,
                          'http://192.168.52.30:3379/v2/keys/abc',
                          'GETTT', None, None, None)
        self.assertRaises(etcd.EtcdRequestError, c.write,
                          dir=True, value='val', key='key')

    def test_etcdincompleteread_exception(self):

        res = etcd.Response(body='abc')
        c = etcd.Client(host=HOSTS)
        self.assertRaises(etcd.EtcdIncompleteRead,
                          c._to_dict, res)

    def test_etcdsslerror_exception(self):

        c = etcd.Client(host=HOSTS)
        self.assertRaises(etcd.EtcdSSLError, etcd.Client,
                          host=HOSTS, protocol='https')
        self.assertRaises(etcd.EtcdSSLError,
                          c._parse_url, 'https://')

    def test_next_server(self):

        utdocker.stop_container('etcd_t0')
        c = etcd.Client(host=HOSTS, read_timeout=2)
        c.set('key', 'val')
        self.assertIn('key', c)

    def test_set_and_get(self):

        cases = (
            ('foo', 'f'),
            ('123', '456'),
            ('12foo', '1f'),
            ('*-@', 'special'),

            ('/bar', 'bb'),
            ('/789', '1000'),
            ('/12bar', '1bb'),
            ('/$*-', 'specialss'),
        )

        cli = etcd.Client(host=HOSTS)
        for key, val in cases:
            cli.set(key, val)
            res = cli.get(key)
            self.assertEqual(val, res.value)

    def test_decode(self):

        cases = (
            ('key1', utfjson.dump(u'我', encoding=None), '"\\u6211"'),

            # when save '"\xb6\xd4"' with etcd but the etcd cannot
            # convert them, so the default '\ufffd\ufffd' was saved.
            # when get it from etcd, '\ufffd\ufffd' was converted into
            # '"\xef\xbf\xbd\xef\xbf\xbd"'.
            ('key2', utfjson.dump(u'对', encoding='gbk'),
             '"\xef\xbf\xbd\xef\xbf\xbd"'),

            ('key3', utfjson.dump(u'我', encoding='utf-8'), '"\xe6\x88\x91"'),

            ('key4', utfjson.dump(u'我'), '"\xe6\x88\x91"'),
            ('key5', utfjson.dump('我'), '"\xe6\x88\x91"'),

            ('key6', utfjson.dump({"我": "我"}),
             '{"\xe6\x88\x91": "\xe6\x88\x91"}'),

            ('key7', utfjson.dump({"我": u"我"}),
             '{"\xe6\x88\x91": "\xe6\x88\x91"}'),

            ('key8', utfjson.dump({u"我": "我"}),
             '{"\xe6\x88\x91": "\xe6\x88\x91"}'),

            ('key9', utfjson.dump({u"我": u"我"}),
             '{"\xe6\x88\x91": "\xe6\x88\x91"}'),

            ('key10', utfjson.dump((u"我",)), '["\xe6\x88\x91"]'),
        )

        cli = etcd.Client(host=HOSTS)
        for key, val, expected in cases:
            cli.set(key, val)
            res = cli.get(key)
            self.assertEqual(expected, res.value)

    def test_ttl(self):

        c = etcd.Client(host=HOSTS)

        c.set('abc', 'val', ttl=1)
        self.assertIn('abc', c)
        time.sleep(2)
        self.assertNotIn('abc', c)

    def test_get_conditions(self):

        c = etcd.Client(host=HOSTS)

        kv = (
            ('key1', 'val1', False),
            ('key2', 'val2', False),
            ('dir1', None, True),
            ('dir1/key3', 'val3', False),
        )
        for k, v, d in kv:
            c.test_and_set(k, v, dir=d)

        nodes = c.get('', sorted=True)._children
        self.assertEqual('/dir1', nodes[0]['key'])
        self.assertEqual('/key1', nodes[1]['key'])
        self.assertEqual('/key2', nodes[2]['key'])

        nodes = c.get('', sorted=True, recursive=False)._children
        self.assertNotIn('nodes', nodes[0])
        nodes = c.get('', sorted=True, recursive=True)._children
        self.assertIn('nodes', nodes[0])

    def test_set_conditions(self):

        c = etcd.Client(host=HOSTS)

        c.test_and_set('foo', 'val')
        c.test_and_set('foo', 'val', prevExist=True)
        self.assertRaises(etcd.EcodeNodeExist, c.test_and_set,
                          'foo', 'val', prevExist=False)

        res = c.test_and_set('foo', 'val1', prevValue='val')
        self.assertEqual('val1', res.value)
        self.assertRaises(etcd.EcodeTestFailed, c.test_and_set,
                          'foo', 'val1', prevValue='val1000')

        res = c.test_and_set('foo', 'val2', prevIndex=res.modifiedIndex)
        self.assertEqual('val2', res.value)

        res = c.test_and_set('foo', None, refresh=True, ttl=2)
        self.assertEqual(2, res.ttl)

    def test_set_append(self):

        c = etcd.Client(host=HOSTS)

        c.test_and_set('dd', None, dir=True)
        c.test_and_set('dd', 'val1', append=True)
        c.test_and_set('dd', 'val2', append=True)

        res = c.get('dd')
        self.assertEqual(2, len(res._children))
        for n in res.leaves:
            self.assertIn(n.value, ['val1', 'val2'])

    def test_watch(self):

        c = etcd.Client(host=HOSTS)

        res = c.set('abc', 'val')
        self.assertEqual(res.value, c.watch(
            'abc', waitindex=res.modifiedIndex).value)
        self.assertRaises(etcd.EtcdReadTimeoutError, c.watch,
                          'abc', waitindex=res.modifiedIndex + 1, timeout=1)

        start_index = res.modifiedIndex
        for i in range(4):
            res = c.set('abc', 'val%d' % (i))
            end_index = res.modifiedIndex

        result = []
        for res in c.eternal_watch('abc', waitindex=start_index + 1,
                                   until=end_index):
            result.append(res.value)

        self.assertListEqual(['val0', 'val1', 'val2', 'val3'], result)

    def test_update(self):

        c = etcd.Client(host=HOSTS)

        res = c.set('key', 'val')
        res.value += 'abc'
        res = c.update(res)

        self.assertEqual('valabc', res.value)

    def test_delete(self):

        c = etcd.Client(host=HOSTS)
        kv = (
            ('key1', 'val1', False),
            ('dir1', None, True),
            ('dir1/key2', 'val2', False),
        )
        for k, v, d in kv:
            c.test_and_set(k, v, dir=d)

        c.delete('key1')
        self.assertNotIn('key1', c)

        self.assertRaises(etcd.EcodeNotFile, c.delete, 'dir1')
        self.assertRaises(etcd.EcodeDirNotEmpty, c.delete, 'dir1', dir=True)

        c.delete('dir1', recursive=True, dir=True)
        self.assertNotIn('dir1', c)

    def test_mkdir(self):

        c = etcd.Client(host=HOSTS)

        res = c.mkdir('d')
        self.assertTrue(res.dir)
        self.assertEqual('/d', res.key)
        self.assertIn('d', c)

    def test_refresh(self):

        c = etcd.Client(host=HOSTS)

        c.set('key', 'val')
        res = c.refresh('key', ttl=1)
        self.assertEqual(1, res.ttl)

    def test_lsdir(self):

        c = etcd.Client(host=HOSTS)

        c.test_and_set('dir1', None, dir=True)
        res = c.lsdir('dir1')
        self.assertTrue(res.dir)
        self.assertEqual('/dir1', res.key)

    def test_rlsdir(self):

        c = etcd.Client(host=HOSTS)

        c.test_and_set('dir1', None, dir=True)
        c.test_and_set('dir1/key', 'val')
        res = c.rlsdir('dir1')
        self.assertTrue(res.dir)
        self.assertEqual('/dir1', res.key)
        self.assertEqual(1, len(res._children))
        self.assertEqual('val', res._children[0]['value'])

    def test_deldir(self):

        c = etcd.Client(host=HOSTS)

        c.test_and_set('dir1', None, dir=True)
        c.deldir('dir1')
        self.assertNotIn('dir1', c)

    def test_rdeldir(self):

        c = etcd.Client(host=HOSTS)

        c.test_and_set('dir1', None, dir=True)
        c.test_and_set('dir1/key', 'val')
        c.rdeldir('dir1')
        self.assertNotIn('dir1', c)

    def _set_auth(self, hosts):

        root_pwd = '123456'
        c = etcd.Client(host=hosts)
        c.create_root(root_pwd)
        c.enable_auth(root_pwd)
        c.revoke_role_permissions(
            'guest', root_pwd, {"write": ["/*"], "read": ["/*"]})
        c.create_role('rw_role', root_pwd, {"write": ["/*"], "read": ["/*"]})
        c.create_user('rw_user', '654321', root_pwd, ["rw_role"])
        time.sleep(1)

    def test_auth(self):

        hosts = (
            ('192.168.52.90', 3379),
            ('192.168.52.91', 3379),
            ('192.168.52.92', 3379),
        )
        names = ('etcd_au0', 'etcd_au1', 'etcd_au2',)
        nodes = ('node_au0', 'node_au1', 'node_au2',)
        _start_etcd_server(hosts, names, nodes)
        time.sleep(10)

        try:
            self._set_auth(hosts)
            c = etcd.Client(host=hosts)
            self.assertRaises(etcd.EcodeInscientPermissions,
                              c.set, 'key', 'val')
            self.assertRaises(etcd.EcodeInscientPermissions, c.get, '')

            c = etcd.Client(
                host=hosts, basic_auth_account='rw_user:654321')
            c.set('key', 'val')
            self.assertIn('key', c)
            res = c.get('key')
            self.assertEqual('val', res.value)

            c.disable_auth('123456')
            c = etcd.Client(host=hosts)
            c.set('key1', 'val1')
            self.assertIn('key1', c)
            res = c.get('key1')
            self.assertEqual('val1', res.value)
        finally:
            utdocker.remove_container(*names)

    def test_st_leader(self):

        c = etcd.Client(host=HOSTS)
        leader = c.st_leader
        self.assertIn('leader', leader)
        self.assertEqual(4, len(leader['followers']))

    def test_st_store(self):

        c = etcd.Client(host=HOSTS)
        store = c.st_store

        expected_res = [
            'getsSuccess',
            'getsFail',
            'setsSuccess',
            'setsFail',
            'deleteSuccess',
            'deleteFail',
            'updateSuccess',
            'updateFail',
            'createSuccess',
            'createFail',
            'compareAndSwapSuccess',
            'compareAndSwapFail',
            'compareAndDeleteSuccess',
            'compareAndDeleteFail',
            'expireCount',
            'watchers',
        ]

        self.assertEqual(set(expected_res), set(store.keys()))

    def test_st_self(self):

        c = etcd.Client(host=HOSTS)
        s = c.st_self

        self.assertTrue(type(s) is dict)
        self.assertTrue(len(s) > 0)

    def test_exception(self):

        c = etcd.Client(host=HOSTS)
        self.assertRaises(etcd.EcodeKeyNotFound, c._st, '/XXX')

    def test_names(self):

        c = etcd.Client(host=HOSTS)
        expected_res = [
            'node_0',
            'node_1',
            'node_2',
            'node_3',
            'node_4',
        ]

        self.assertEqual(set(expected_res), set(c.names))

    def test_ids(self):

        c = etcd.Client(host=HOSTS)
        self.assertEqual(5, len(c.ids))

    def test_clienturls(self):

        expected_res = ['http://%s:%d' % (ip, port) for ip, port in HOSTS]
        c = etcd.Client(host=HOSTS)
        self.assertEqual(set(expected_res), set(c.clienturls))

    def test_peerurls(self):

        expected_res = ['http://%s:3380' % (ip) for ip, _ in HOSTS]
        c = etcd.Client(host=HOSTS)
        self.assertEqual(set(expected_res), set(c.peerurls))

    def test_change_peerurls(self):

        c = etcd.Client(host=HOSTS)
        mem = c.members[0]
        old_urls = mem['peerURLs']
        new_urls = [old_urls[0] + '1']

        c.change_peerurls(mem['id'], *new_urls)
        self.assertIn(new_urls[0], c.peerurls)

        c.change_peerurls(mem['id'], *old_urls)
        self.assertIn(old_urls[0], c.peerurls)

        self.assertRaises(etcd.EtcdException, c.change_peerurls, mem['id'], [])

    def test_del_member(self):

        hosts = (
            ('192.168.52.50', 3379),
            ('192.168.52.51', 3379),
            ('192.168.52.52', 3379),
        )
        names = ('etcd_d0', 'etcd_d1', 'etcd_d2')
        nodes = ('node_d0', 'node_d1', 'node_d2',)
        _start_etcd_server(hosts, names, nodes)

        time.sleep(10)
        c = etcd.Client(host=hosts, read_timeout=1)

        try:
            for i in range(2):
                ids = c.ids
                c.del_member(ids[0])
                self.assertEqual(2 - i, len(c.members))
                time.sleep(10)
        finally:
            utdocker.remove_container(*names)

    def test_add_member(self):

        hosts = [
            ('192.168.52.70', 3379),
            ('192.168.52.71', 3379),
        ]
        names = ['etcd_a0', 'etcd_a1']
        nodes = ['node_a0', 'node_a1']
        _start_etcd_server(hosts, names, nodes)
        time.sleep(10)

        cases = (
            ('192.168.52.72', 'etcd_a2', 'node_a2', 3),
            ('192.168.52.73', 'etcd_a3', 'node_a3', 4),
        )

        c = etcd.Client(host=hosts)
        try:
            for ip, name, node, count in cases:
                c.add_member(*['http://%s:3380' % (ip)])
                hosts.append((ip, 3379))
                nodes.append(node)
                names.append(name)

                utdocker.start_container(name, etcd_test_tag, ip,
                                         _generate_command(
                                             len(hosts) - 1,
                                             hosts,
                                             nodes,
                                             state='existing'))

                time.sleep(10)

                self.assertEqual(count, len(c.members))
        finally:
            utdocker.remove_container(*names)

    def _clear_users_roles(self, root_pwd):

        c = etcd.Client(host=HOSTS)

        users = c.get_user(None, root_pwd)
        if users is not None and users['users'] is not None:
            for u in users['users']:
                try:
                    c.delete_user(u['user'], root_pwd)
                except:
                    continue

        roles = c.get_role(None, root_pwd)
        if roles is not None and roles['roles'] is not None:
            for r in roles['roles']:
                try:
                    c.delete_role(r['role'], root_pwd)
                except:
                    continue

    def test_create_root(self):

        c = etcd.Client(host=HOSTS)
        res = c.create_root('123')

        self.assertEqual('root', res['user'])

    def test_create_and_get_user(self):

        c = etcd.Client(host=HOSTS)

        cases = (
            ('u1', 'p1'),
            ('u2', 'p2'),
            ('u3', 'p3'),
            ('u4', 'p4'),
        )

        all_users = {'users': []}
        try:
            for u, p in cases:
                c.create_user(u, p, '')
                res = c.get_user(u, '')
                self.assertEqual({'user': u}, res)

                all_users['users'].append(res)
                res = c.get_user(None, '')
                self.assertEqual(all_users, res)
        finally:
            self._clear_users_roles('')

    def test_create_and_get_role(self):

        c = etcd.Client(host=HOSTS)

        cases = ('r1', 'r2', 'r3', 'r4')

        # default role cannot delete
        all_roles = [
            {'role': 'root',
             'permissions': {'kv': {'read': ['/*'], 'write': ['/*']}}}
        ]

        try:
            for r in cases:
                c.create_role(r, '')
                res = c.get_role(r, '')
                expected = {
                    'role': r,
                    'permissions': {'kv': {'read': None, 'write': None}}
                }
                self.assertEqual(expected, res)

                all_roles.insert(-1, res)
                res = c.get_role(None, '')
                self.assertEqual(all_roles, res['roles'])
        finally:
            self._clear_users_roles('')

    def test_grant_revoke_user_roles(self):

        c = etcd.Client(host=HOSTS)
        c.create_user('u_test', '11', '')
        c.create_role('r_test', '')

        try:
            c.grant_user_roles('u_test', '', ['r_test'])
            res = c.get_user('u_test', '')
            self.assertEqual('r_test', res['roles'][0]['role'])

            c.revoke_user_roles('u_test', '', ['r_test'])
            res = c.get_user('u_test', '')
            self.assertNotIn('roles', res)
        finally:
            self._clear_users_roles('')

    def test_grant_revoke_role_permissions(self):

        c = etcd.Client(host=HOSTS)
        c.create_role('r_test', '')

        try:
            c.grant_role_permissions('r_test', '', {'read': ['/*']})
            r = c.get_role('r_test', '')
            self.assertEqual(['/*'], r['permissions']['kv']['read'])

            c.revoke_role_permissions('r_test', '', {'read': ['/*']})
            r = c.get_role('r_test', '')
            self.assertEqual([], r['permissions']['kv']['read'])
        finally:
            self._clear_users_roles('')


def _generate_command(index, hosts, nodes, state='new'):

    node = nodes[index]
    ip, _ = hosts[index]
    cluster = ''
    for i in range(len(hosts)):
        cluster += '%s=http://%s:3380,' % (nodes[i], hosts[i][0])

    cluster = cluster[:-1]

    return "--name {name} \
           --initial-advertise-peer-urls http://{ip_peer_adv}:3380 \
           --listen-peer-urls http://{ip_peer}:3380 \
           --advertise-client-urls http://{ip_cli_adv}:3379 \
           --listen-client-urls http://{ip_cli}:3379 \
           --initial-cluster {cluster} \
           --initial-cluster-state {state} \
           --initial-cluster-token test_etcd_token" \
           .format(name=node, ip_peer_adv=ip, ip_peer=ip,
                   ip_cli_adv=ip, ip_cli=ip, cluster=cluster, state=state)


def _start_etcd_server(hosts, names, nodes, state='new'):

    utdocker.create_network()

    for i in range(len(hosts)):

        name = names[i]
        ip, _ = hosts[i]
        command = _generate_command(i, hosts, nodes, state=state)

        utdocker.start_container(name, etcd_test_tag, ip, command)



#!/usr/bin/env python
# coding: utf-8

import unittest

from pykit import redisutil
from pykit import utdocker
from pykit import utfjson
from pykit import ututil
from pykit import zktx
from pykit.zktx.status import COMMITTED

dd = ututil.dd

redis_tag = 'daocloud.io/redis:3.2.3'
redis_port = 6379


class TestRedis(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        utdocker.pull_image(redis_tag)

    def setUp(self):
        utdocker.create_network()

        utdocker.start_container('redis-0', redis_tag, '192.168.52.40', '')
        dd('started redis in docker')

        redisutil.wait_serve(('192.168.52.40', redis_port))

        self.redis_cli = redisutil.get_client(('192.168.52.40', redis_port))
        self.txid_path = "tx/txidset"
        self.storage = zktx.RedisStorage(self.redis_cli, self.txid_path)

    def tearDown(self):
        utdocker.remove_container('redis-0')

    def test_kv_get_set(self):
        accessor = zktx.RedisKeyValue(self.redis_cli)

        cases = (
            ('k1', 1),
            ('k2', {}),
            ('k3', []),
            ('k4', 'foo'),
            ('k5', {"xx": "xx"}),
            ('k6', [1, 2]),
        )

        for k, v in cases:
            accessor.set(k, v)
            actual = accessor.get(k)
            self.assertEqual(v, actual)

    def test_kv_hget_hset(self):
        accessor = zktx.RedisKeyValue(self.redis_cli)

        cases = (
            ('h1', 'k1', 1),
            ('h2', 'k2', {}),
            ('h3', 'k3', []),
            ('h4', 'k4', 'foo'),
            ('h5', 'k5', {"xx": "xx"}),
            ('h6', 'k6', [1, 2]),
        )

        for h, k, v in cases:
            accessor.hset(h, k, v)
            actual = accessor.hget(h, k)
            self.assertEqual(v, actual)

    def test_kv_get_path(self):
        def get_path(k):
            return 'foo/' + k

        accessor = zktx.RedisKeyValue(self.redis_cli, get_path=get_path)
        cases = (
            ('k1', 1),
            ('k2', {}),
            ('k3', []),
            ('k4', 'foo'),
            ('k5', {"xx": "xx"}),
            ('k6', [1, 2]),
        )

        for k, v in cases:
            accessor.set(k, v)
            actual = self.redis_cli.get(get_path(k))
            self.assertEqual(v, utfjson.load(actual))

    def test_v_get_set(self):
        def get_path():
            return 'foo'

        accessor = zktx.RedisValue(self.redis_cli, get_path=get_path)

        cases = (
            1,
            {},
            [],
            'foo',
            {'xx': 'xx'},
            [1, 2],
        )

        for v in cases:
            accessor.set(v)
            actual = accessor.get()
            self.assertEqual(v, actual)

    def test_load(self):
        def load(val):
            return 'foo'

        accessor = zktx.RedisKeyValue(self.redis_cli, load=load)
        accessor.set('xx', 'xx')
        val = accessor.get('xx')

        self.assertEqual('foo', val)

    def test_get_txidset(self):
        cases = (
            [[1, 2]],
            [[7, 20]],
            [[1, 3], [7, 8]],
            [[1, 2], [10, 15]],
        )

        for c in cases:
            val = {
                COMMITTED: c,
            }
            self.redis_cli.delete(self.txid_path)
            self.redis_cli.set(self.txid_path, utfjson.dump(val))
            txidset = self.storage.txidset.get()
            self.assertEqual(c, txidset[COMMITTED])

    def test_apply_jour(self):
        cases = (
            {'meta/h1/k1': 1},
            {'meta/h2/k2': {}},
            {'meta/h3/k3': []},
            {'meta/h4/k4': 'foo'},
            {'meta/h5/k5': {'xx': 'xxx'}},

            {'meta/h10/k10': 1, 'meta/h100/k100': 10},
            {'meta/h11/k11': {}, 'meta/h101/k101': {}},
            {'meta/h12/k12': [], 'meta/h102/k102': []},
            {'meta/h13/k13': 'foo', 'meta/h103/k103': 'bar'},
            {'meta/h14/k14': {'xx': 'xxx'}, 'meta/h104/k104': {'yy': 'yyy'}},
        )

        for jour in cases:
            self.storage.apply_jour(jour)

            for k, v in jour.items():
                k_parts = k.split('/', 2)

                actual = self.redis_cli.hget(k_parts[1], k_parts[2])

                self.assertEqual(v, utfjson.load(actual))

    def test_set_txidset(self):
        cases = (
            [[1, 2]],
            [[7, 20]],
            [[1, 3], [7, 8]],
            [[1, 2], [10, 15]],
        )

        for c in cases:
            self.storage.set_txidset(COMMITTED, c)

            txidset = self.storage.txidset.get()
            self.assertEqual(c, txidset[COMMITTED])



#!/usr/bin/env python2
# coding: utf-8

import copy
import logging
import Queue

import MySQLdb

logger = logging.getLogger(__name__)


retriable_err = (2006, 2013)


class MysqlConnectionPool(object):

    def __init__(self, conn_argkw, options=None):

        options = options or {}
        options = copy.deepcopy(options)

        self.options = options
        self.queue = Queue.Queue(32)
        self.conn_argkw = conn_argkw
        if 'host' in conn_argkw:
            self.name = '{host}:{port}'.format(**conn_argkw)
        else:
            self.name = '{unix_socket}'.format(**conn_argkw)

        self.stat = {'name': self.name,
                     'create': 0,
                     'pool_get': 0,
                     'pool_put': 0,
                     }

    def __call__(self, action=None):

        if action is None or action == 'get_conn':
            return ConnectionWrapper(self)
        elif action == 'stat':
            return self.stat
        else:
            raise ValueError(action, 'invalid action: ' + repr(action))

    def get_conn(self):

        try:
            conn = self.queue.get(block=False)
            self.stat['pool_get'] += 1
            logger.debug('reuse connection:' + repr(self.stat))
            return conn

        except Queue.Empty:

            conn = new_connection(self.conn_argkw, options=self.options)

            self.stat['create'] += 1
            logger.info('create new connection: ' + repr(self.stat))

            return conn

    def put_conn(self, conn):

        try:
            self.queue.put(conn, block=False)
            self.stat['pool_put'] += 1
            logger.debug('put connection:' + repr(self.stat))

        except Queue.Full:
            conn.close()

    def query(self, sql, use_dict=True, retry=0):

        if retry < 0:
            retry = 0

        retry = int(retry)

        # the first attempt does not count as 'retry'
        for i in range(retry + 1):

            try:
                with self() as conn:
                    return conn_query(conn, sql, use_dict=use_dict)

            except MySQLdb.OperationalError as e:
                if len(e.args) > 0 and e[0] in retriable_err:
                    logger.info(
                        repr(e) + " conn_query error {sql}".format(sql=sql))
                    continue
                else:
                    raise
        else:
            raise


class ConnectionWrapper(object):

    def __init__(self, pool):
        self.pool = pool
        self.conn = None

    def __enter__(self):
        self.conn = self.pool.get_conn()
        return self.conn

    def __exit__(self, errtype, errval, _traceback):

        if errtype is None:
            self.pool.put_conn(self.conn)
            self.conn = None
        else:
            self.conn.close()


def make(conn_argkw, options=None):

    return MysqlConnectionPool(conn_argkw, options=options)


def conn_query(conn, sql, use_dict=True):

    if use_dict:
        cur = conn.cursor(MySQLdb.cursors.DictCursor)
    else:
        cur = conn.cursor()

    cur.execute(sql)
    rst = cur.fetchall()
    cur.close()

    return rst


def new_connection(conn_argkw, options=None):

    # useful arg could be added in future.:
    # conn_argkw.init_command

    options = options or {}

    opt = {
        'autocommit': 1,
    }
    opt.update(options)

    conn = MySQLdb.connect(**conn_argkw)
    for k, v in opt.items():
        conn.query('set {k}={v}'.format(k=k, v=v))

    return conn



from .jobq import (
    EmptyRst,
    Finish,
    run,
    stat,
    JobManager,

    JobWorkerError,
    JobWorkerNotFound,
)

from .works import (
    limit_job_speed,
)

__all__ = [
    'EmptyRst',
    'Finish',
    'run',
    'stat',
    'JobManager',

    'JobWorkerError',
    'JobWorkerNotFound',
    'limit_job_speed',
]



#!/usr/bin/env python2
# coding: utf-8

import logging
import threading
import time
from collections import OrderedDict

import psutil
from geventwebsocket import Resource
from geventwebsocket import WebSocketApplication
from geventwebsocket import WebSocketError
from geventwebsocket import WebSocketServer

from pykit import threadutil
from pykit import utfjson
from pykit import jobq

logger = logging.getLogger(__name__)

MEM_AVAILABLE = 'mem_available'
CPU_IDLE_PERCENT = 'cpu_idle_percent'
CLIENT_NUMBER = 'client_number'
JOBS_DIR = 'jobs'

CHECK_LOAD_PARAMS = {
    'mem_low_threshold': {
        'load_name': MEM_AVAILABLE,
        'default': 500 * 1024 ** 2,  # 500M
        'greater': True,
    },
    'cpu_low_threshold': {
        'load_name': CPU_IDLE_PERCENT,
        'default': 3,  # 3%
        'greater': True,
    },
    'max_client_number': {
        'load_name': CLIENT_NUMBER,
        'default': 1000,
        'greater': False,
    },
}


class SystemOverloadError(Exception):
    pass


class JobError(Exception):
    pass


class InvalidMessageError(JobError):
    pass


class InvalidProgressError(InvalidMessageError):
    pass


class LoadingError(JobError):
    pass


class JobNotInSessionError(JobError):
    pass


class Job(object):

    lock = threading.RLock()
    sessions = {}

    def __init__(self, channel, msg, func):

        self.ident = msg['ident']
        self.channel = channel
        self.data = msg
        self.worker = func
        self.ctx = {}
        self.err = None

        if self.ident in self.sessions:
            logger.info('job: %s already exists, created by chennel %s' %
                        (self.ident, repr(self.sessions[self.ident].channel)))
            return
        else:
            self.sessions[self.ident] = self
            logger.info(('inserted job: %s to sessions by channel %s, ' +
                         'there are %d jobs in sessions now') %
                        (self.ident, repr(self.channel), len(self.sessions)))

        self.thread = threadutil.start_thread(target=self.work, args=(),
                                              daemon=True)

    def work(self):

        logger.info("job %s started, the data is: %s" %
                    (self.ident, self.data))

        try:
            self.worker(self)
        except Exception as e:
            logger.exception('job %s got exception: %s' %
                             (self.ident, repr(e)))
            self.err = e
        finally:
            logger.info('job %s ended' % self.ident)
            self.close()

    def close(self):

        with self.lock:
            del self.sessions[self.ident]
            logger.info(('removed job: %s from sessions, there are %d ' +
                         'jobs in sessions now') %
                        (self.ident, len(self.sessions)))


def get_or_create_job(channel, msg, func):

    with Job.lock:
        Job(channel, msg, func)

        job = Job.sessions.get(msg['ident'])

    return job


def progress_sender(job, channel, interval=5, stat=None):

    stat = stat or (lambda data: data)
    data = job.data

    i = 10
    try:
        while True:
            # if thread died due to some reason, still send 10 stats
            if not job.thread.is_alive():
                logger.info('job %s died: %s' % (job.ident, repr(job.err)))
                if i == 0:
                    channel.ws.close()
                    break
                i -= 1

            logger.info('jod %s on channel %s send progress: %s' %
                        (job.ident, repr(channel), repr(stat(data))))

            to_send = stat(data)
            if channel.report_system_load and type(to_send) == type({}):
                to_send['system_load'] = channel.get_system_load()

            channel.ws.send(utfjson.dump(to_send))

            time.sleep(interval)

    except WebSocketError as e:
        if channel.ws.closed == True:
            logger.info('the client has closed the connection')
        else:
            logger.exception(('got websocket error when sending progress on' +
                              ' channel %s: %s') % (repr(channel), repr(e)))

    except Exception as e:
        logger.exception('got exception when sending progress on channel %s: %s'
                         % (repr(channel), repr(e)))
        channel.ws.close()


class JobdWebSocketApplication(WebSocketApplication):

    jobq_mgr = None

    def on_open(self):
        logger.info('on open, the channel is: ' + repr(self))
        self.ignore_message = False

    def _parse_request(self, message):
        try:
            try:
                msg = utfjson.load(message)
            except Exception as e:
                raise InvalidMessageError(
                    'message is not a vaild json string: %s' % message)

            self._check_msg(msg)

            self.report_system_load = msg.get('report_system_load') == True
            self.cpu_sample_interval = msg.get('cpu_sample_interval', 0.02)

            if not isinstance(self.cpu_sample_interval, (int, long, float)):
                raise InvalidMessageError(
                    'cpu_sample_interval is not a number')

            check_load = msg.get('check_load')
            if type(check_load) == type({}):
                self._check_system_load(check_load)

            self.jobs_dir = msg.get('jobs_dir', JOBS_DIR)

            self._setup_response(msg)
            return

        except SystemOverloadError as e:
            logger.info('system overload on chennel %s, %s'
                        % (repr(self), repr(e)))
            self._send_err_and_close(e)

        except JobError as e:
            logger.info('error on channel %s while handling message, %s'
                        % (repr(self), repr(e)))
            self._send_err_and_close(e)

        except Exception as e:
            logger.exception(('exception on channel %s while handling ' +
                              'message, %s') % (repr(self), repr(e)))
            self._send_err_and_close(e)

    def on_message(self, message):
        logger.info('on message, the channel is: %s, the message is: %s' %
                    (repr(self), message))
        if self.ignore_message:
            return

        else:
            self.ignore_message = True

        self.jobq_mgr.put((self, message))

    def _send_err_and_close(self, err):
        try:
            err_msg = {
                'err': err.__class__.__name__,
                'val': err.args,
            }
            self.ws.send(utfjson.dump(err_msg))
        except Exception as e:
            logger.error(('error on channel %s while sending back error '
                          + 'message, %s') % (repr(self), repr(e)))

    def get_system_load(self):
        return {
            MEM_AVAILABLE: psutil.virtual_memory().available,
            CPU_IDLE_PERCENT: psutil.cpu_times_percent(
                self.cpu_sample_interval).idle,
            CLIENT_NUMBER: len(self.protocol.server.clients),
        }

    def _check_system_load(self, check_load):
        system_load = self.get_system_load()

        for param_name, param_attr in CHECK_LOAD_PARAMS.iteritems():
            param_value = check_load.get(param_name, param_attr['default'])

            if not isinstance(param_value, (int, long, float)):
                raise InvalidMessageError('%s is not a number' % param_name)

            load_name = param_attr['load_name']
            diff = system_load[load_name] - param_value

            if not param_attr['greater']:
                diff = 0 - diff

            if diff < 0:
                raise SystemOverloadError(
                    '%s: %d is %s than: %d' %
                    (load_name, system_load[load_name],
                     param_attr['greater'] and 'less' or 'greater',
                     param_value))

    def _check_msg(self, msg):
        if type(msg) != type({}):
            raise InvalidMessageError("message is not dictionary")

        if 'ident' not in msg:
            raise InvalidMessageError("'ident' is not in message")

        if 'func' not in msg:
            raise InvalidMessageError("'func' is not in message")

    def _setup_response(self, msg):

        func = self._get_func_by_name(msg)
        channel = self
        job = get_or_create_job(channel, msg, func)

        if job is None:
            raise JobNotInSessionError(
                'job not in sessions: ' + repr(Job.sessions))

        progress = msg.get('progress', {})

        if progress in (None, False):
            return

        if type(progress) != type({}):
            raise InvalidProgressError(
                'the progress in message is not a dictionary')

        interval = progress.get('interval', 5)
        progress_key = progress.get('key')

        if progress_key is None:
            lam = lambda r: r
        else:
            lam = lambda r: r[progress_key]

        threadutil.start_thread(target=progress_sender,
                                args=(job, channel, interval, lam),
                                daemon=True)

    def _get_func_by_name(self, msg):
        mod_func = self.jobs_dir.split('/') + msg['func'].split('.')
        mod_path = '.'.join(mod_func[:-1])
        func_name = mod_func[-1]

        try:
            mod = __import__(mod_path)
        except (ImportError, SyntaxError) as e:
            raise LoadingError('failed to import %s: %s' % (mod_path, repr(e)))

        for mod_name in mod_path.split('.')[1:]:
            mod = getattr(mod, mod_name)

        logger.info('mod imported from: ' + repr(mod.__file__))

        try:
            func = getattr(mod, func_name)
        except AttributeError as e:
            raise LoadingError("function not found: " + repr(func_name))

        return func

    def on_close(self, reason):
        logger.info('on close, the channel is: ' + repr(self))


def _parse_request(args):
    app, msg = args
    app._parse_request(msg)


def run(ip='127.0.0.1', port=63482, jobq_thread_count=10):
    JobdWebSocketApplication.jobq_mgr = jobq.JobManager(
        [(_parse_request, jobq_thread_count)])

    WebSocketServer(
        (ip, port),
        Resource(OrderedDict({'/': JobdWebSocketApplication})),
    ).serve_forever()




import time
import unittest

from kazoo.client import KazooClient
from kazoo.exceptions import ConnectionClosedError

from pykit import config
from pykit import threadutil
from pykit import utdocker
from pykit import ututil
from pykit import zkutil

dd = ututil.dd

zk_test_name = 'zk_test'
zk_test_tag = 'daocloud.io/zookeeper:3.4.10'
zk_test_auth = ('digest', 'xp', '123')
zk_test_acl = (('xp', '123', 'cdrw'), )

# zookeeper docker env vars:
# https://hub.docker.com/_/zookeeper/
#
# Example stack.yml for zookeeper:
#
# version: '3.1'
# services:
#   zoo1:
#     image: zookeeper
#     restart: always
#     hostname: zoo1
#     ports:
#       - 2181:2181
#     environment:
#       ZOO_MY_ID: 1
#       ZOO_SERVERS: server.1=0.0.0.0:2888:3888 server.2=zoo2:2888:3888 server.3=zoo3:2888:3888
#   zoo2: ...
#   zoo3: ...
#
# ZOO_TICK_TIME          : Defaults to 2000. ZooKeeper's tickTime
# ZOO_INIT_LIMIT         : Defaults to 5. ZooKeeper's initLimit
# ZOO_SYNC_LIMIT         : Defaults to 2. ZooKeeper's syncLimit
# ZOO_MAX_CLIENT_CNXNS   : Defaults to 60. ZooKeeper's maxClientCnxns
# ZOO_STANDALONE_ENABLED : Defaults to false. Zookeeper's standaloneEnabled
# ZOO_MY_ID
# ZOO_SERVERS
#
# Where to store data
#       This image is configured with volumes at /data and /datalog to hold the Zookeeper in-memory database snapshots and the transaction log of updates to the database, respectively.


class TestZKLock(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        utdocker.pull_image(zk_test_tag)

    def setUp(self):

        config.zk_acl = zk_test_acl
        config.zk_auth = zk_test_auth

        self.counter = 0
        self.running = True

        utdocker.create_network()
        utdocker.start_container(
            zk_test_name,
            zk_test_tag,
            env={
                "ZOO_MY_ID": 1,
                "ZOO_SERVERS": "server.1=0.0.0.0:2888:3888",
            },
            port_bindings={
                2181: 2181,
            }
        )

        self.zk = KazooClient(hosts='127.0.0.1:2181')
        self.zk.start()
        scheme, name, passw = zk_test_auth
        self.zk.add_auth(scheme, name + ':' + passw)

        # create lock base dir
        acl = zkutil.make_kazoo_digest_acl(zk_test_acl)
        self.zk.create('lock/', acl=acl)

        self.lck = zkutil.ZKLock('foo_name', zkclient=self.zk)

    def tearDown(self):

        self.zk.stop()
        utdocker.remove_container(zk_test_name)

    def _on_conn_change(self, state):

        self.lsn_count += 1

    def test_bounded_listener(self):

        # ensure that adding a bounded listener(self.on_xxx) is ok

        self.lsn_count = 0

        self.zk.add_listener(self._on_conn_change)
        self.zk.add_listener(self._on_conn_change)

        self.zk.stop()

        self.assertEqual(1, self.lsn_count)

    def _loop_acquire(self, n, ident):

        zk = KazooClient(hosts='127.0.0.1:2181')
        zk.start()
        scheme, name, passw = zk_test_auth
        zk.add_auth(scheme, name + ':' + passw)

        for ii in range(n):
            l = zkutil.ZKLock('foo_name', zkclient=zk)
            with l:

                self.total += 1
                self.counter += 1

                self.assertTrue(self.counter == 1)

                time.sleep(0.01)
                self.counter -= 1

                dd("id={ident:0>2} n={ii:0>2} got and released lock: {holder}".format(
                    ident=ident,
                    ii=ii,
                    holder=l.lock_holder))

        zk.stop()

    def test_concurrent(self):

        self.running = True
        self.total = 0
        n_repeat = 40
        n_thread = 5

        ths = []
        for ii in range(n_thread):
            t = threadutil.start_daemon_thread(self._loop_acquire, args=(n_repeat, ii,))
            ths.append(t)

        for th in ths:
            th.join()

        self.running = False
        self.assertEqual(n_repeat * n_thread, self.total)

    def test_persistent(self):
        l = zkutil.ZKLock('foo_name', ephemeral=False, on_lost=lambda: True)
        try:
            with l:
                l.zkclient.stop()
        except ConnectionClosedError:
            pass

        self.assertRaises(zkutil.LockTimeout, self.lck.acquire, timeout=0.2)

    def test_timeout(self):

        l1 = zkutil.ZKLock('foo_name', on_lost=lambda: True)
        l2 = zkutil.ZKLock('foo_name', on_lost=lambda: True)

        with l1:
            with ututil.Timer() as t:
                self.assertRaises(zkutil.LockTimeout, l2.acquire, timeout=0.2)
                self.assertAlmostEqual(0.2, t.spent(), places=1)

            with ututil.Timer() as t:
                self.assertRaises(zkutil.LockTimeout, l2.acquire, timeout=-1)
                self.assertAlmostEqual(0.0, t.spent(), delta=0.01)

        try:
            l2.acquire(timeout=-1)
        except zkutil.LockTimeout:
            self.fail('timeout<0 should could acquire')

    def test_lock_holder(self):
        a = zkutil.ZKLock('foo_name', on_lost=lambda: True)
        b = zkutil.ZKLock('foo_name', on_lost=lambda: True)

        with a:
            self.assertEqual((a.identifier, 0), a.lock_holder)
            val, zstate = self.zk.get(a.lock_path)
            self.assertEqual((val, zstate.version), a.lock_holder)

            locked, holder, ver = b.try_acquire()
            self.assertFalse(locked)
            self.assertEqual((a.identifier, 0), (holder, ver))
            self.assertEqual((val, zstate.version), (holder, ver))

    def test_watch_acquire(self):

        a = zkutil.ZKLock('foo', on_lost=lambda: True)
        b = zkutil.ZKLock('foo', on_lost=lambda: True)

        # no one locked

        n = 0
        for holder, ver in a.acquire_loop():
            n += 1
        self.assertEqual(0, n, 'acquired directly')

        # watch node change

        it = b.acquire_loop()

        holder, ver = it.next()
        self.assertEqual((a.identifier, 0), (holder, ver))

        self.zk.set(a.lock_path, 'xx')
        holder, ver = it.next()
        self.assertEqual(('xx', 1), (holder, ver), 'watched node change')

        a.release()
        try:
            holder, ver = it.next()
            self.fail('should not have next yield')
        except StopIteration:
            pass

        self.assertTrue(b.is_locked())

    def test_try_lock(self):

        l1 = zkutil.ZKLock('foo_name', on_lost=lambda: True)
        l2 = zkutil.ZKLock('foo_name', on_lost=lambda: True)

        with l1:
            with ututil.Timer() as t:
                locked, holder, ver = l2.try_acquire()
                self.assertFalse(locked)
                self.assertEqual(l1.identifier, holder)
                self.assertGreaterEqual(ver, 0)

                self.assertAlmostEqual(0.0, t.spent(), delta=0.05)

        with ututil.Timer() as t:
            locked, holder, ver = l2.try_acquire()
            self.assertTrue(locked)
            self.assertEqual(l2.identifier, holder)
            self.assertEqual(ver, 0)

            self.assertAlmostEqual(0.0, t.spent(), delta=0.05)

    def test_try_release(self):

        l1 = zkutil.ZKLock('foo_name', on_lost=lambda: True)
        l2 = zkutil.ZKLock('foo_name', on_lost=lambda: True)

        released, holder, ver = l1.try_release()
        self.assertEqual((True, l1.identifier, -1), (released, holder, ver))

        with l2:
            released, holder, ver = l1.try_release()
            self.assertEqual((False, l2.identifier, 0), (released, holder, ver))

            released, holder, ver = l2.try_release()
            self.assertEqual((True, l2.identifier, 0), (released, holder, ver))

    def test_zk_lost(self):

        sess = {'acquired': True}

        def watch(state):
            dd('zk node state changed to: ', state)
            sess['acquired'] = False

        self.zk.add_listener(watch)

        # test zk close

        l = zkutil.ZKLock('foo_name', zkclient=self.zk)

        l.acquire()
        self.zk.stop()
        time.sleep(0.1)
        self.assertFalse(sess['acquired'])

        # test node delete

        sess['acquired'] = True
        self.zk.start()
        self.zk.add_listener(watch)

        l = zkutil.ZKLock('foo_name', zkclient=self.zk)

        with l:
            time.sleep(0.1)
            self.zk.delete(l.zkconf.lock('foo_name'))
            time.sleep(0.1)
            self.assertFalse(sess['acquired'])

    def test_node_change_after_acquired(self):

        sess = {'acquired': True}

        def on_lost():
            sess['acquired'] = False

        l = zkutil.ZKLock('foo_name',
                          zkclient=self.zk,
                          on_lost=on_lost)

        with l:
            sess['acquired'] = True
            self.zk.delete(l.zkconf.lock('foo_name'))
            time.sleep(0.1)
            self.assertFalse(sess['acquired'])

        l = zkutil.ZKLock('foo_name',
                          zkclient=self.zk,
                          on_lost=on_lost)

        with l:
            sess['acquired'] = True
            self.zk.set(l.zkconf.lock('foo_name'), 'xxx')
            time.sleep(0.1)
            self.assertFalse(sess['acquired'])

    def test_node_change_after_released(self):

        sess = {'acquired': True}

        def on_lost():
            sess['acquired'] = False

        l = zkutil.ZKLock('foo_name',
                          zkclient=self.zk,
                          on_lost=on_lost)

        with l:
            sess['acquired'] = True

        time.sleep(0.1)
        self.assertTrue(sess['acquired'])

    def test_is_locked(self):

        l = zkutil.ZKLock('foo_name', zkclient=self.zk)

        with l:
            pass

        self.assertFalse(l.is_locked())

        l = zkutil.ZKLock('foo_name', zkclient=self.zk)
        l.acquire()
        self.assertTrue(l.is_locked())
        l.try_release()
        self.assertFalse(l.is_locked())

    def test_conn_lost_when_blocking_acquiring(self):

        l2 = zkutil.ZKLock('foo_name', on_lost=lambda: True)

        th = threadutil.start_daemon(target=self.zk.stop, after=0.5)
        with l2:
            try:
                self.lck.acquire(timeout=1)
                self.fail('expected connection error')
            except ConnectionClosedError:
                pass

        th.join()

    def test_internal_zkclient(self):

        sess = {'acquired': True}

        def on_lost():
            sess['acquired'] = False

        # There must be a listener specified to watch connection issue
        self.assertRaises(ValueError, zkutil.ZKLock, 'foo_name')

        l = zkutil.ZKLock('foo_name', on_lost=on_lost)

        with l:
            self.zk.delete(l.zkconf.lock('foo_name'))
            time.sleep(0.1)
            self.assertFalse(sess['acquired'])

    def test_acl(self):
        with self.lck:
            acls, zstat = self.zk.get_acls(self.lck.lock_path)

        dd(acls)
        self.assertEqual(1, len(acls))

        acl = acls[0]
        expected = zkutil.perm_to_long(zk_test_acl[0][2], lower=False)

        self.assertEqual(set(expected), set(acl.acl_list))
        self.assertEqual('digest', acl.id.scheme)
        self.assertEqual(zk_test_acl[0][0], acl.id.id.split(':')[0])

    def test_config(self):

        old = (config.zk_acl, config.zk_auth, config.zk_node_id)

        config.zk_acl = (('foo', 'bar', 'cd'),
                         ('xp', '123', 'cdrwa'))

        config.zk_auth = ('digest', 'xp', '123')
        config.zk_node_id = 'abc'

        l = zkutil.ZKLock('foo_name', on_lost=lambda: True)

        dd(l.zkconf.acl())

        def _check_ac(ac):
            self.assertEqual('digest', ac.id.scheme)
            self.assertEqual('foo', ac.id.id.split(':')[0])
            self.assertEqual(set(['CREATE', 'DELETE']), set(ac.acl_list))

        _check_ac(l.zkconf.kazoo_digest_acl()[0])

        with l:
            # should have created lock node
            data, zstate = self.zk.get(l.lock_path)
            dd(data)

            self.assertEqual('abc', data.split('-')[0])

            acls, zstate = self.zk.get_acls(l.lock_path)
            dd(acls)

            _check_ac(acls[0])

        (config.zk_acl, config.zk_auth, config.zk_node_id) = old

    def test_hosts(self):

        l = zkutil.ZKLock('foo_name',
                          zkconf=dict(
                              hosts='127.0.0.1:2181',
                          ),
                          on_lost=lambda: True)

        with l:
            self.assertEqual('127.0.0.1:2181', l._hosts)

    def test_specify_identifier(self):

        a = zkutil.ZKLock('foo_name',
                          zkconf=dict(
                              hosts='127.0.0.1:2181',
                          ),
                          identifier='faked',
                          on_lost=lambda: True)

        b = zkutil.ZKLock('foo_name',
                          zkconf=dict(
                              hosts='127.0.0.1:2181',
                          ),
                          identifier='faked',
                          on_lost=lambda: True)

        a.acquire()
        b.acquire()
        dd('a and b has the same identifier thus they both can acquire the lock')

    def test_release_listener_removed(self):

        self.lck.release()
        self.assertNotIn(self.lck.on_connection_change, self.zk.state_listeners)

    def test_release_owning_client_stopped(self):

        l = zkutil.ZKLock('foo_name',
                          zkconf=dict(
                              hosts='127.0.0.1:2181',
                          ),
                          on_lost=lambda: True)

        l.release()
        self.assertTrue(l.zkclient._stopped.is_set())



from .block import (
    BlockID,
    BlockIDError,
)

from .block_index import (

    BlockIndexError,

    BlockIndex,
)

from .block_group import (
    BlockGroup,
    BlockNotFoundError,
    BlockTypeNotSupported,
    BlockTypeNotSupportReplica,
)

from .block_group_id import (
    BlockGroupIDError,

    BlockGroupID,
)

from .server import (
    get_serverrec_str,
    idc_distance,
    make_serverrec,
    validate_idc,

    ServerID,
    DriveID,

    DriveIDError,
)

from .region import (
    Region,

    BlockNotInRegion,
)

__all__ = [
    "BlockID",
    "BlockIDError",

    "BlockIndexError",
    "BlockIndex",

    "BlockGroup",
    "BlockGroupID",
    "BlockGroupIDError",
    "BlockNotFoundError",
    "BlockTypeNotSupported",
    "BlockTypeNotSupportReplica",

    "get_serverrec_str",
    "idc_distance",
    "make_serverrec",
    "validate_idc",

    "ServerID",
    "DriveID",

    "DriveIDError",

    "Region",

    "BlockNotInRegion",
]



import threading
import unittest

from pykit import threadutil
from pykit import txutil
from pykit import ututil

dd = ututil.dd


class MyError(Exception):
    pass


class TestCASLoop(unittest.TestCase):

    def setUp(self):
        self.lock = threading.RLock()
        self.val = 0
        self.ver = 0

    def tearDown(self):
        pass

    def _get(self):
        with self.lock:
            return self.val, self.ver

    def _set(self, val, prev_stat):
        with self.lock:
            if prev_stat != self.ver:
                raise txutil.CASConflict(prev_stat, self.ver)
            self.val = val
            self.ver += 1

    def _set_raise_myerror(self, val, prev_stat):
        with self.lock:
            raise MyError(prev_stat, self.ver)

    def test_cas(self):

        i = 0
        for curr in txutil.cas_loop(self._get, self._set):
            self.assertEqual(i, curr.stat)
            curr.v += 2

    def test_cas_n(self):

        i = 0
        for curr in txutil.cas_loop(self._get, self._set_raise_myerror,
                                    conflicterror=MyError):
            self.assertEqual(i, curr.n)
            i += 1
            if i == 5:
                break

    def test_cas_abort(self):

        for curr in txutil.cas_loop(self._get, self._set):
            curr.v += 2
            break

        self.assertEqual((0, 0), (self.val, self.ver))

    def test_cas_customed_error(self):

        for curr in txutil.cas_loop(self._get, self._set_raise_myerror,
                                    conflicterror=MyError):
            curr.v += 2
            if curr.n == 5:
                break

        self.assertEqual((0, 0), (self.val, self.ver))

    def test_cas_argument(self):

        rst = {}

        def _get(*args, **kwargs):
            rst['get'] = args, kwargs
            return 'val', 'stat'

        def _set(*args, **kwargs):
            rst['set'] = args, kwargs

        for curr in txutil.cas_loop(_get, _set, ('foo', 'bar'), dict(a=1, b=2)):
            curr.v = 'uservalue'

        self.assertEqual((('foo', 'bar'), {"a": 1, "b": 2}), rst['get'])
        self.assertEqual((('foo', 'bar', 'uservalue', 'stat'), {"a": 1, "b": 2}), rst['set'])

        # empty arg

        rst = {}

        for curr in txutil.cas_loop(_get, _set):
            curr.v = 'uservalue'

        self.assertEqual(((), {}), rst['get'])
        self.assertEqual((('uservalue', 'stat'), {}), rst['set'])

    def test_cas_concurrent(self):

        def _update():
            for ii in range(10):
                for curr in txutil.cas_loop(self._get, self._set):
                    curr.v += 1

        ths = [threadutil.start_daemon(_update)
               for _ in range(10)]

        for th in ths:
            th.join()

        self.assertEqual((100, 100), (self.val, self.ver))



#!/usr/bin/env python2
# coding: utf-8

import errno
import logging
import os

import subprocess32

logger = logging.getLogger(__name__)


class ProcError(Exception):

    def __init__(self, returncode, out, err, cmd, arguments, options):

        super(ProcError, self).__init__(returncode,
                                        out,
                                        err,
                                        cmd,
                                        arguments,
                                        options)

        self.returncode = returncode
        self.out = out
        self.err = err
        self.command = cmd
        self.arguments = arguments
        self.options = options


def command(cmd, *arguments, **options):

    close_fds = options.get('close_fds', True)
    cwd = options.get('cwd', None)
    env = options.get('env', None)
    if env is not None:
        env = dict(os.environ, **env)
    stdin = options.get('stdin', None)

    subproc = subprocess32.Popen([cmd] + list(arguments),
                                 close_fds=close_fds,
                                 cwd=cwd,
                                 env=env,
                                 stdin=subprocess32.PIPE,
                                 stdout=subprocess32.PIPE,
                                 stderr=subprocess32.PIPE, )

    out, err = subproc.communicate(input=stdin)

    subproc.wait()

    return (subproc.returncode, out, err)


def command_ex(cmd, *arguments, **options):
    returncode, out, err = command(cmd, *arguments, **options)
    if returncode != 0:
        raise ProcError(returncode, out, err, cmd, arguments, options)

    return returncode, out, err


def shell_script(script_str, **options):
    options['stdin'] = script_str
    return command('sh', **options)


def _waitpid(pid):
    while True:
        try:
            os.waitpid(pid, 0)
            break
        except OSError as e:
            # In case we encountered an OSError due to EINTR (which is
            # caused by a SIGINT or SIGTERM signal during
            # os.waitpid()), we simply ignore it and enter the next
            # iteration of the loop, waiting for the child to end.  In
            # any other case, this is some other unexpected OS error,
            # which we don't want to catch, so we re-raise those ones.
            if e.errno != errno.EINTR:
                raise


def _close_fds():
    try:
        max_fd = os.sysconf("SC_OPEN_MAX")
    except ValueError:
        max_fd = 65536

    for i in range(max_fd):
        try:
            os.close(i)
        except OSError:
            pass


def start_process(cmd, target, env, *args):

    try:
        pid = os.fork()
    except OSError as e:
        logger.error(repr(e) + ' while fork')
        raise

    if pid == 0:
        _close_fds()
        args = list(args)
        env = dict(os.environ, **env)
        args.append(env)
        os.execlpe(cmd, cmd, target, *args)
    else:
        _waitpid(pid)




import logging
import uuid

logger = logging.getLogger(__name__)

# try to load config from a top-level module.
try:
    import pykitconfig
except ImportError:
    pykitconfig = object()
    logger.info('pykitconfig not found by "import pykitconfig".'
                ' Using default config.'
                ' You can create file "pykitconf.py" to define default config for pykit.')


def _get(key, default=None):
    v = getattr(pykitconfig, key, default)

    logger.debug('set pykit config: {key}={v}'.format(key=key, v=v))
    return v


uid                  = _get('uid')
gid                  = _get('gid')
log_dir              = _get('log_dir')
cat_stat_dir         = _get('cat_stat_dir')
iostat_stat_path     = _get('iostat_stat_path', '/tmp/pykit-iostat')
zk_acl               = _get('zk_acl')                   # (('xp', '123', 'cdrwa'), ('foo', 'bar', 'rw'))
zk_auth              = _get('zk_auth')                  # ('digest', 'xp', '123')
zk_hosts             = _get('zk_hosts', '127.0.0.1:2181')
zk_lock_dir          = _get('zk_lock_dir', 'lock/')
zk_node_id           = _get('zk_node_id', '%012x' % uuid.getnode())
zk_record_dir        = _get('zk_record_dir', 'record/')
zk_tx_dir            = _get('zk_tx_dir', 'tx/')
zk_seq_dir           = _get('zk_seq_dir', 'seq/')
rp_cli_nwr           = _get('rp_cli_nwr', (3, 2, 2))
rp_cli_ak_sk         = _get('rp_cli_ak_sk', ('access_key', 'secret_key'))



#!/usr/bin/env python2
# coding: utf-8

import json
import mimetypes
import os

OCTET_STREAM = 'application/octet-stream'

dir_name = os.path.dirname(__file__)
resource = os.path.join(dir_name, 'thirdpart', 'mimes.json')


def _init():
    with open(resource, 'r') as f:
        content = f.read()
    return json.loads(content)


mimes = _init()


def get_by_filename(filename):
    mime_type = None

    if filename.find('.') != -1:
        suffix = filename.rsplit('.', 1)[-1]
        mime_type = mimes.get(suffix)
        if mime_type is None:
            mime_type, _ = mimetypes.guess_type(filename)

    return mime_type or OCTET_STREAM



#!/usr/bin/env python2
# coding: utf-8

import shelve
import time

from pykit import config

from . import fsutil

MAJOR = 0
MINOR = 1
NAME = 2

READ_N = 3
READ_MERGED_N = 4
READ_SEC_N = 5
READ_MS = 6

WRITE_N = 7
WRITE_MERGED_N = 8
WRITE_SEC_N = 9
WRITE_MS = 10

IO_INPROC_N = 11
IO_MS = 12
IO_WEIGHTED_MS = 13

tmpl = {
    "name": NAME,
    "r": {
        "n": READ_N,
        "merge_n": READ_MERGED_N,
        "sec_n": READ_SEC_N,
        "ms": READ_MS,
    },
    "w": {
        "n": WRITE_N,
        "merge_n": WRITE_MERGED_N,
        "sec_n": WRITE_SEC_N,
        "ms": WRITE_MS,
    },
    "io": {
        "inproc_n": IO_INPROC_N,
        "ms": IO_MS,
        "weighted_ms": IO_WEIGHTED_MS,
    },
}


class DeviceNotFound(Exception):
    pass


def iostat(device=None, path=None, stat_path=None):

    # device: /dev/sdb*
    # path:   /home/guest
    # devname: sdb*
    #
    # returns {
    #   ioutil: 0-100 percentage
    #   read: Byte/s
    #   write: Byte/s
    # }

    assert device is not None or path is not None

    if stat_path is None:
        stat_path = config.iostat_stat_path

    if device is None:
        device = fsutil.get_device(path)

    devname = device.split('/', 2)[2]

    while True:
        curr = load_st(devname)
        prev = read_prev(stat_path, devname)

        if prev is None:
            write_prev(stat_path, devname, curr)
            time.sleep(1)
            continue

        duration = curr['ts'] - prev['ts']
        if curr['ts'] - prev['ts'] <= 0:
            write_prev(stat_path, devname, curr)
            time.sleep(1)
            continue

        if duration > 20:
            write_prev(stat_path, devname, curr)

        break

    st = {'ioutil': curr['io']['ms'] - prev['io']['ms'],
          'read':  curr['r']['byte'] - prev['r']['byte'],
          'write': curr['w']['byte'] - prev['w']['byte'],
          }
    for k, v in st.items():
        st[k] = v / duration

    # ioutil is recorded in millisecond
    st['ioutil'] = st['ioutil'] * 100 / 1000
    st['ioutil'] = min([100, st['ioutil']])

    return st


def read_prev(stat_path, devname):

    d = None
    try:
        d = shelve.open(stat_path)
        return d.get(devname)

    finally:
        if d is not None:
            d.close()


def write_prev(stat_path, devname, st):

    d = None
    try:
        d = shelve.open(stat_path)
        d[devname] = st

    finally:
        if d is not None:
            d.close()


def load_st(devname):

    sys_key = get_sys_key(devname)

    sts = fsutil.read_file('/proc/diskstats')
    lines = sts.strip().split("\n")

    elts = []
    for line in lines:
        elts = line.strip().split()
        if elts[NAME] == sys_key:
            break
    else:
        raise DeviceNotFound("device not found in /proc/diskstats: " + sys_key)

    sector_size = fsutil.read_file('/sys/block/{sys_key}/queue/hw_sector_size'.format(
        sys_key=sys_key))

    sector_size = int(sector_size)

    st = format_dev_st(elts, sector_size)
    st['name'] = devname
    return st


def get_sys_key(devname):

    if devname.startswith('mapper/'):
        # device on lvm.
        # /proc/diskstats use name "dm-*" for lvm disk
        lvs = get_all_lvs()
        n = lvs.index('/dev/' + devname)
        return 'dm-' + str(n)

    if devname.startswith('cciss/'):
        # cciss/c0d0p1
        return devname

    # NOTE: on some old OS, there is no 14 column for partition like
    # "/dev/sdb1", which is found on "218.30.114.74".
    if not devname.startswith('loop'):
        # //dev/loop2 has no parent dev
        return devname.rstrip('0123456789')

    return devname


def get_all_lvs():

    lvs = []

    maps = fsutil.read_file('/etc/mtab').strip()
    maps = maps.split('\n')

    for m in maps:
        m = m.strip()
        if m == "":
            continue

        nm = m.split()[0]
        if nm.startswith('/dev/mapper/'):
            lvs.append(nm)

    return lvs


def format_dev_st(elts, sector_size):

    st = {}

    for k, i in tmpl.items():
        if isinstance(i, int):
            if elts[i].isdigit():
                st[k] = int(elts[i])
            else:
                st[k] = elts[i]
        else:
            st[k] = {}
            for k2, j in tmpl[k].items():
                st[k][k2] = int(elts[j])

    st['ts'] = int(time.time())
    st['r']['byte'] = st['r']['sec_n'] * sector_size
    st['w']['byte'] = st['w']['sec_n'] * sector_size
    return st



#!/usr/bin/env python2
# coding: utf-8

import types

K = 1024 ** 1
M = 1024 ** 2
G = 1024 ** 3
T = 1024 ** 4
P = 1024 ** 5
E = 1024 ** 6
Z = 1024 ** 7
Y = 1024 ** 8

value_to_unit = {
    1: '',
    K: 'K',
    M: 'M',
    G: 'G',
    T: 'T',
    P: 'P',
    E: 'E',
    Z: 'Z',
    Y: 'Y',
}

unit_to_value = dict([(v, k)
                      for (k, v) in value_to_unit.items()
                      if v != ''
                      ])

integer_types = (type(0), type(0L))


def humannum_int(i, unit=None):

    i = int(i)

    if unit is None:

        unit = K

        # keep at least 2 digit
        while i / unit > 0:
            unit *= K

        unit /= K

        while unit not in value_to_unit:
            unit /= K

    v = i * 1.0 / unit

    if v == int(v):
        return '%d%s' % (v, value_to_unit[unit])

    if v > 10:
        vlen = 1
    elif v > 1:
        vlen = 2
    else:
        vlen = 3

    return ('%.' + str(vlen) + 'f%s') % (v, value_to_unit[unit])


def humannum(data, unit=None, include=None, exclude=None):

    if isinstance(data, types.DictType):

        data = data.copy()

        keys = set(data.keys())
        if include is not None:
            keys = keys & set(include)

        if exclude is not None:
            keys = keys - set(exclude)

        for k in keys:
            data[k] = humannum(data[k])

        return data

    elif isinstance(data, types.BooleanType):
        # We have to deal with bool because for historical reason bool is
        # subclass of int.
        # When bool is introduced into python 2.2 it is represented with int,
        # similar to C.
        return data

    elif isinstance(data, types.ListType):
        return [humannum(x) for x in data]

    elif isinstance(data, types.StringTypes):
        return data

    elif isinstance(data, integer_types):
        return humannum_int(data, unit=unit)

    elif isinstance(data, types.FloatType):
        if data > 999:
            return humannum_int(int(data), unit=unit)
        elif abs(data) < 0.0000000001:
            return '0'
        else:
            return '%.2f' % (data)

    else:
        return data


def parseint(data, safe=None):
    return int(parsenum(data, safe=safe))


def parsenum(data, safe=None):

    if safe is None:
        safe = False

    original = data

    if isinstance(data, integer_types):
        return data

    if not isinstance(data, types.StringTypes):
        return data

    if data == '':
        return 0

    if data.endswith('%'):
        fl = float(data[:-1]) / 100.0
        return fl

    data = data.upper().rstrip('B').rstrip('I')

    unit_name = data[-1]

    try:
        if unit_name in unit_to_value:
            val = float(data[:-1]) * unit_to_value[unit_name]
        else:
            val = float(data)

        if val == int(val):
            val = int(val)

    except ValueError:
        if safe:
            val = original
        else:
            raise

    return val



#!/usr/bin/env python2.6
# coding: utf-8

import unittest

from pykit import utfjson


class TestUTFJson(unittest.TestCase):

    def test_load(self):

        self.assertEqual(None, utfjson.load(None))
        self.assertEqual({}, utfjson.load('{}'))

        # load unicode, result in utf-8

        self.assertEqual('我',      utfjson.load('"\\u6211"'))
        self.assertEqual(str,  type(utfjson.load('"\\u6211"')))

        # unicode and string in a dictionary.

        obj = '{"a": "\\u6211", "b": "1"}'
        rst = utfjson.load(obj)

        self.assertEqual({"a": "\xe6\x88\x91", "b": "1"}, rst)
        self.assertEqual(str, type(rst["a"]))
        self.assertEqual(str, type(rst["b"]))

        # load utf-8, result in str

        rst = utfjson.load('"\xe6\x88\x91"')
        self.assertEqual('我', rst)
        self.assertEqual(str,  type(rst))

        # load gbk, result in str, in gbk encoding

        gbk = '"\xb6\xd4\xd5\xbd\xc6\xbd\xcc\xa8\xb9\xd9\xb7\xbd\xd7\xee\xd0\xc2\xb0\xe6"'
        self.assertEqual('对战平台官方最新版'.decode('utf-8').encode('gbk'), utfjson.load(gbk))
        self.assertEqual(str, type(utfjson.load(gbk)))

        # load any

        s = '"\xbb"'
        rst = utfjson.load(s)
        self.assertEqual('\xbb', rst)
        self.assertEqual(str, type(rst))

    def test_load_backslash_x_encoded(self):

        s = '"\\x61"'
        self.assertEqual('a', utfjson.load(s))

        s = '"\\X61"'
        self.assertEqual('a', utfjson.load(s))

        s = '"\\xe6\\x88\\x91"'
        self.assertEqual('我', utfjson.load(s))

        self.assertRaises(utfjson.JSONDecodeError, utfjson.load, '"\\"')
        self.assertRaises(utfjson.JSONDecodeError, utfjson.load, '"\\x"')
        self.assertRaises(utfjson.JSONDecodeError, utfjson.load, '"\\x6"')

    def test_load_decode(self):
        self.assertEqual('我', utfjson.load('"我"'))
        self.assertEqual(u'我', utfjson.load('"我"', encoding='utf-8'))
        self.assertEqual(unicode, type(utfjson.load('"我"', encoding='utf-8')))

        self.assertEqual({'a': u"我".encode('utf-8')}, utfjson.load('{"a": "\\u6211"}'))
        self.assertEqual({'a': u"我"}, utfjson.load('{"a": "我"}', encoding='utf-8'))
        self.assertEqual({'a':  "我"}, utfjson.load('{"a": "我"}'))
        self.assertEqual({'a':  "我"}, utfjson.load('{"a": "我"}'))
        self.assertEqual(["我"],       utfjson.load('["我"]'))


    def test_dump(self):

        self.assertEqual('null', utfjson.dump(None))
        self.assertEqual('{}', utfjson.dump({}))

        self.assertRaises(TypeError, utfjson.dump, '我',      encoding=None)
        self.assertRaises(TypeError, utfjson.dump, {'我':1},  encoding=None)
        self.assertRaises(TypeError, utfjson.dump, {1:'我'},  encoding=None)
        self.assertRaises(TypeError, utfjson.dump, ['我'],    encoding=None)
        self.assertRaises(TypeError, utfjson.dump, [('我',)], encoding=None)

        self.assertEqual('"\\u6211"',      utfjson.dump(u'我', encoding=None))
        self.assertEqual('"\xb6\xd4"',     utfjson.dump(u'对', encoding='gbk'))
        self.assertEqual('"\xe6\x88\x91"', utfjson.dump(u'我', encoding='utf-8'))

        self.assertEqual('"\xe6\x88\x91"', utfjson.dump(u'我'))
        self.assertEqual('"\xe6\x88\x91"', utfjson.dump( '我'))

        # by default unicode are encoded

        self.assertEqual('{"\xe6\x88\x91": "\xe6\x88\x91"}', utfjson.dump({ "我":  "我"}))
        self.assertEqual('{"\xe6\x88\x91": "\xe6\x88\x91"}', utfjson.dump({ "我": u"我"}))
        self.assertEqual('{"\xe6\x88\x91": "\xe6\x88\x91"}', utfjson.dump({u"我":  "我"}))
        self.assertEqual('{"\xe6\x88\x91": "\xe6\x88\x91"}', utfjson.dump({u"我": u"我"}))
        self.assertEqual('["\xe6\x88\x91"]',                 utfjson.dump((u"我", )))

        self.assertEqual('{"\\u6211": "\\u6211"}', utfjson.dump({u"我": u"我"}, encoding=None))

        self.assertEqual('"\\""', utfjson.dump('"'))

        # encoded chars and unicode chars in one string
        self.assertEqual('/aaa\xe7\x89\x88\xe6\x9c\xac/jfkdsl\x01', utfjson.load('"\/aaa\xe7\x89\x88\xe6\x9c\xac\/jfkdsl\u0001"'))



import logging

from pykit import logutil

logger = logutil.make_logger('/tmp', log_fn='stdlog', level='INFO',
                             fmt='message')


logutil.add_std_handler(logger, 'stdout', fmt='message', level=logging.ERROR)
logger.debug('debug')
logger.info('stdlog')
logger.error('error')



import os

import yaml

dir_name = os.path.dirname(__file__)
file_name = os.path.join(dir_name, 'error_code.yaml')

with open(file_name, 'r') as f:
    file_content = f.read()

error_code = yaml.safe_load(file_content)



import base64
import datetime
import hmac
import json
import urllib
from hashlib import sha256

SIGV4_TIMESTAMP = '%Y%m%dT%H%M%SZ'
CREDENTIALS_SUFFIX = 'aws4_request'
EMPTY_PAYLOAD_HASH = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
ALGORITHM = 'AWS4-HMAC-SHA256'
UNSIGNED_PAYLOAD = 'UNSIGNED-PAYLOAD'
VALID_AUTH_ARGS = [
    'X-Amz-Algorithm',
    'X-Amz-Credential',
    'X-Amz-Date',
    'X-Amz-Expires',
    'X-Amz-SignedHeaders',
    'X-Amz-Signature',
]


class InvalidSigningDateError(Exception):
    pass


class InvalidHeadersNotToSignError(Exception):
    pass


class RequestDateFormatError(Exception):
    pass


class InvalidRequestError(Exception):
    pass


class InvalidRequestVerbError(InvalidRequestError):
    pass


class InvalidRequestArgError(InvalidRequestError):
    pass


class InvalidRequestUriError(InvalidRequestError):
    pass


class InvalidRequestHeadersError(InvalidRequestError):
    pass


class InvalidRequestArgNameError(InvalidRequestArgError):
    pass


class InvalidRequestArgValueError(InvalidRequestArgError):
    pass


class InvalidRequestHeaerNameError(InvalidRequestHeadersError):
    pass


class SigV4(object):

    def _to_utf8(self, string):
        if isinstance(string, unicode):
            string = string.encode('utf-8')

        return string

    def _escape(self, s, safe='/'):
        s = self._to_utf8(s)
        return urllib.quote(s, safe)

    def _unescape_plus(self, s):
        s = self._to_utf8(s)
        return urllib.unquote_plus(s)

    def _uri_encode_args(self, args):
        encoded_args = {}
        for k, v in args.items():
            encoded_k = self._escape(k, '~')

            if isinstance(v, basestring):
                encoded_v = self._escape(v, '~')
            elif type(v) == type([]):
                encoded_v = []
                for value in v:
                    if isinstance(value, basestring):
                        encoded_v.append(self._escape(value, '~'))
                    else:
                        encoded_v.append(value)
            else:
                encoded_v = v

            encoded_args[encoded_k] = encoded_v

        return encoded_args

    def _build_canonical_query_string(self, encoded_args):
        arg_names = []
        for k, _ in encoded_args.items():
            if k != 'X-Amz-Signature':
                arg_names.append(k)

        arg_names.sort()

        key_value_strs = []
        for name in arg_names:
            value = encoded_args[name]

            if type(value) == type([]):
                value = value[0]

            if not isinstance(value, str):
                value = ''

            key_value_strs.append(name + '=' + value)

        return '&'.join(key_value_strs)

    def _utf8_join(self, string_list, separator):
        utf8_list = []
        for string in string_list:
            utf8_list.append(self._to_utf8(string))

        return separator.join(utf8_list)

    def _build_canonical_headers(self, signed_headers, headers):
        l = []

        for name in signed_headers.split(';'):
            value = headers[name]

            if type(value) == type([]):
                value_str = self._utf8_join(value, ',')
            elif isinstance(value, basestring):
                value_str = self._to_utf8(value)
            else:
                value_str = ''

            l.append(name + ':' + value_str)

        return '\n'.join(l) + '\n'

    def _build_canonical_request(self, ctx):
        return '\n'.join([
            ctx['verb'],
            ctx['uri'],
            ctx['canonical_query_string'],
            ctx['canonical_headers'],
            ctx['signed_headers'],
            ctx['hashed_payload'],
        ])

    def _build_string_to_sign(self, ctx):
        return '\n'.join([
            ctx['algorithm'],
            ctx['request_date'],
            ctx['credential_scope'],
            ctx['hashed_canonical_request'],
        ])

    def _make_sha256(self, s):
        return sha256(s).hexdigest()

    def _make_hmac_sha256(self, key, msg, to_hex=False):
        if to_hex:
            sig = hmac.new(key, msg, sha256).hexdigest()
        else:
            sig = hmac.new(key, msg, sha256).digest()
        return sig

    def _derive_signing_key(self, secret_key, credential_scope):
        scope_items = credential_scope.split('/')

        k_date = self._make_hmac_sha256('AWS4' + secret_key, scope_items[0])
        k_region = self._make_hmac_sha256(k_date, scope_items[1])
        k_service = self._make_hmac_sha256(k_region, scope_items[2])
        k_signing = self._make_hmac_sha256(k_service, scope_items[3])

        return k_signing

    def _calc_signature(self, signing_key, string_to_sign):
        return self._make_hmac_sha256(signing_key, string_to_sign, to_hex=True)

    def _query_string_to_args(self, query_string):
        args = {}

        if query_string == '':
            return args

        items = query_string.split('&')

        for item in items:
            key, eq_sign, value = item.partition('=')
            arg_name = self._unescape_plus(key)

            if arg_name not in args:
                args[arg_name] = []
            if eq_sign == '=':
                args[arg_name].append(self._unescape_plus(value))
            else:
                args[arg_name].append(True)

        for arg_name, arg_value in args.items():
            if len(arg_value) == 1:
                args[arg_name] = arg_value[0]

        return args

    def _args_to_query_string(self, args):
        qs = []
        encoded_args = self._uri_encode_args(args)

        for arg_name, arg_value in encoded_args.items():
            if isinstance(arg_value, list):
                for value in arg_value:
                    if isinstance(value, str):
                        qs.append(arg_name + '=' + value)
                    elif value == True:
                        qs.append(arg_name)

            elif isinstance(arg_value, str):
                qs.append(arg_name + '=' + arg_value)

            elif arg_value:
                qs.insert(0, arg_name)

        return '&'.join(qs)


class Signer(SigV4):

    def __init__(self, access_key, secret_key, region=None, service=None, default_expires=None):
        self._access_key = self._to_utf8(access_key)
        self._secret_key = self._to_utf8(secret_key)
        self._service = self._to_utf8(service) or 's3'
        self._region = self._to_utf8(region) or 'us-east-1'
        self._default_expires = default_expires or 60

    def _get_request_date(self, request_date):
        if isinstance(request_date, (int, long, float)):
            dt = datetime.datetime.utcfromtimestamp(request_date)
            request_date = dt.strftime(SIGV4_TIMESTAMP)
        elif isinstance(request_date, basestring):
            try:
                datetime.datetime.strptime(request_date, SIGV4_TIMESTAMP)
            except:
                raise RequestDateFormatError(
                    'request date: %s is not iso base formmat like 20161206T120102Z' % str(request_date))
        else:
            datetime_now = datetime.datetime.utcnow()
            request_date = datetime_now.strftime(SIGV4_TIMESTAMP)

        return request_date

    def _clean_query_string(self, query_string):
        items = query_string.split('&')
        qs = []

        for item in items:
            arg_name = item.partition('=')[0]
            arg_name = self._unescape_plus(arg_name)

            if arg_name in VALID_AUTH_ARGS:
                continue

            qs.append(item)

        return '&'.join(qs)

    def _clean_args(self, args):
        for arg_name in VALID_AUTH_ARGS:
            if arg_name in args:
                args.pop(arg_name)

        return args

    def _trimall(self, s):
        return ' '.join(s.split())

    def _standardize_headers(self, headers, headers_not_to_sign):
        headers_not_to_sign_low = []
        for h_name in headers_not_to_sign:
            h_name = self._to_utf8(h_name)
            headers_not_to_sign_low.append(h_name.lower())

        stand_headers = {}

        for k, v in headers.items():
            k = self._to_utf8(k)
            v = self._to_utf8(v)

            low_name = k.lower().strip()
            stand_v = self._trimall(str(v))

            if low_name not in stand_headers:
                stand_headers[low_name] = []

            stand_headers[low_name].append(stand_v)

        for k, v in stand_headers.items():
            if len(v) == 1:
                stand_headers[k] = v[0]

        signed_header_names = []

        for h_name in stand_headers.keys():
            if h_name not in headers_not_to_sign_low:
                signed_header_names.append(h_name)

        signed_header_names.sort()

        return ';'.join(signed_header_names), stand_headers

    def _validate_arg_value(self, arg_value):
        if arg_value == True or isinstance(arg_value, basestring):
            return

        if not isinstance(arg_value, list):
            raise InvalidRequestArgValueError(
                'arg value: %s, must be string or list or True' % repr(arg_value))

        for value in arg_value:
            if not isinstance(value, basestring) and value != True:
                raise InvalidRequestArgValueError(
                    'multi arg value: %s, must be string or True' % repr(value))

    def _validate_uri_and_args(self, uri, args):
        if not isinstance(uri, basestring) or not uri.startswith('/'):
            raise InvalidRequestUriError(
                'uri: %s, must be a string and starts with /' % repr(uri))

        has_query_string = False
        if len(uri.split('?')) > 1:
            has_query_string = True
        if has_query_string == True and args != None:
            raise InvalidRequestError(
                'use both query string and args is not allowed')

        if args == None:
            return

        if type(args) != type({}):
            raise InvalidRequestArgError('args: %s, is not a dict' % repr(args))

        for arg_name, arg_value in args.items():
            if not isinstance(arg_name, basestring):
                raise InvalidRequestArgNameError(
                    'arg name: %s, is not a string' % repr(arg_name))

            self._validate_arg_value(arg_value)

    def _validate_headers(self, headers):
        if not isinstance(headers, dict):
            raise InvalidRequestHeadersError(
                'headers: %s, is not a dict' % repr(headers))

        has_host = False
        for k, _ in headers.items():
            if not isinstance(k, basestring):
                raise InvalidRequestHeaerNameError(
                    'header name: %s, is not a string' % repr(k))

            if k.lower() == 'host':
                has_host = True

        if has_host != True:
            raise InvalidRequestHeadersError('absence of host header')

    def _validate_request(self, request):
        if not isinstance(request, dict):
            raise InvalidRequestError(
                'request: %s, is not a dict' % repr(request))
        if not isinstance(request.get('verb'), basestring):
            raise InvalidRequestVerbError('absence of or invalid request verb')

        self._validate_uri_and_args(request.get('uri'), request.get('args'))

        self._validate_headers(request.get('headers'))

    def _modify_request_headers(self, request, query_auth,
                                sign_payload, request_date):
        has_amz_date = False
        hashed_payload = None

        for k, v in request['headers'].items():
            low_name = k.lower()

            if low_name == 'authorization':
                request['headers'].pop(k)
            elif low_name == 'x-amz-date':
                request['headers'].pop(k)
                has_amz_date = True
            elif low_name == 'x-amz-content-sha256':
                hashed_payload = self._to_utf8(v)
                request['headers'].pop(k)

        if has_amz_date == True or query_auth != True:
            request['headers']['X-Amz-Date'] = request_date

        if hashed_payload != None:
            request['headers']['X-Amz-Content-SHA256'] = hashed_payload
            return hashed_payload

        if query_auth == True:
            return UNSIGNED_PAYLOAD

        if sign_payload != True:
            request['headers']['X-Amz-Content-SHA256'] = UNSIGNED_PAYLOAD
            return UNSIGNED_PAYLOAD

        if isinstance(request.get('body'), basestring) and len(request['body']) > 0:
            hashed_payload = self._make_sha256(self._to_utf8(request['body']))
        else:
            hashed_payload = EMPTY_PAYLOAD_HASH

        request['headers']['X-Amz-Content-SHA256'] = hashed_payload
        return hashed_payload

    def _get_credential_scope(self, request_date, signing_date=None):
        credential_date = signing_date or request_date[:8]

        if not isinstance(credential_date, basestring):
            raise InvalidSigningDateError(
                'invalid signing date: %s, not of string type' % repr(credential_date))

        credential_scope = '/'.join([
            credential_date,
            self._region,
            self._service,
            CREDENTIALS_SUFFIX
        ])

        return credential_scope

    def add_auth(self, request, **argkv):
        self._validate_request(request)

        query_auth = argkv.get('query_auth', False)
        sign_payload = argkv.get('sign_payload', False)
        headers_not_to_sign = argkv.get('headers_not_to_sign', [])

        if not isinstance(headers_not_to_sign, list):
            raise InvalidHeadersNotToSignError(
                'argument headers_not_to_sign is not a list')

        request_date = self._get_request_date(
            self._to_utf8(argkv.get('request_date')))

        credential_scope = self._get_credential_scope(
            request_date, self._to_utf8(argkv.get('signing_date')))

        credential = self._access_key + '/' + credential_scope

        hashed_payload = self._modify_request_headers(
            request, query_auth, sign_payload, request_date)

        signed_headers, stand_headers = self._standardize_headers(
            request['headers'], headers_not_to_sign)

        utf8_uri = self._to_utf8(request['uri'])
        origin_uri_path, delimiter, origin_query_string = utf8_uri.partition(
            '?')

        if delimiter != '?':
            origin_query_string = None

        ctx = {
            'verb': self._to_utf8(request['verb']),
            'uri': self._escape(self._unescape_plus(origin_uri_path), '/~'),
            'algorithm': ALGORITHM,
            'request_date': request_date,
            'credential_scope': credential_scope,
            'signed_headers': signed_headers,
            'hashed_payload': hashed_payload,
        }

        if origin_query_string != None:
            cleaned_origin_query_string = self._clean_query_string(
                origin_query_string)
            args = self._query_string_to_args(cleaned_origin_query_string)
        else:
            args = self._clean_args(request.get('args', {}))
            query_string_from_args = self._args_to_query_string(args)

        auth_args = {}
        if query_auth == True:
            amz_expires = str(argkv.get('expires') or self._default_expires)
            amz_expires = self._to_utf8(amz_expires)
            auth_args = {
                'X-Amz-Algorithm': ALGORITHM,
                'X-Amz-Credential': credential,
                'X-Amz-Date': request_date,
                'X-Amz-Expires': amz_expires,
                'X-Amz-SignedHeaders': ctx['signed_headers'],
            }

            for k, v in auth_args.items():
                args[k] = v

        encoded_args = self._uri_encode_args(args)

        ctx['canonical_query_string'] = self._build_canonical_query_string(
            encoded_args)

        ctx['canonical_headers'] = self._build_canonical_headers(
            signed_headers, stand_headers)

        ctx['canonical_request'] = self._build_canonical_request(ctx)

        ctx['hashed_canonical_request'] = self._make_sha256(
            ctx['canonical_request'])

        ctx['string_to_sign'] = self._build_string_to_sign(ctx)

        ctx['signing_key'] = self._derive_signing_key(
            self._secret_key, credential_scope)

        ctx['signature'] = self._calc_signature(
            ctx['signing_key'], ctx['string_to_sign'])

        if origin_query_string != None:
            qs = origin_query_string
        else:
            qs = query_string_from_args

        if query_auth == True:
            if len(qs) > 0:
                qs += '&'
            qs += self._args_to_query_string(auth_args)
            qs += '&X-Amz-Signature=' + ctx['signature']

            request['uri'] = origin_uri_path + '?' + qs
        else:
            if len(qs) != 0:
                request['uri'] = origin_uri_path + '?' + qs
            else:
                request['uri'] = origin_uri_path

            l = ['AWS4-HMAC-SHA256 Credential=%s' % credential]
            l.append('SignedHeaders=%s' % ctx['signed_headers'])
            l.append('Signature=%s' % ctx['signature'])
            request['headers']['Authorization'] = ', '.join(l)

        return ctx

    def add_post_auth(self, fields, **argkv):
        request_date = self._get_request_date(
            self._to_utf8(argkv.get('request_date')))

        credential_scope = self._get_credential_scope(
            request_date, self._to_utf8(argkv.get('signing_date')))

        credential = self._access_key + '/' + credential_scope

        ctx = {
            'algorithm': ALGORITHM,
            'request_date': request_date,
            'credential_scope': credential_scope,
            'policy': fields.get('Policy'),
        }

        policy_json_str = json.dumps(ctx['policy'])
        ctx['string_to_sign'] = base64.b64encode(policy_json_str)

        ctx['signing_key'] = self._derive_signing_key(
            self._secret_key, credential_scope)

        ctx['signature'] = self._calc_signature(
            ctx['signing_key'], ctx['string_to_sign'])

        fields['Policy'] = ctx['string_to_sign']
        fields['X-Amz-Algorithm'] = ALGORITHM
        fields['X-Amz-Credential'] = credential
        fields['X-Amz-Date'] = request_date
        fields['X-Amz-Signature'] = ctx['signature']

        return ctx



import logging
import os

from pykit import fsutil
from pykit.cgrouparch import cgroup_util

logger = logging.getLogger(__name__)


def set_cgroup(cgroup_path, conf):
    if 'weight' in conf:
        weight_file = os.path.join(cgroup_path, 'blkio.weight')

        weight = str(conf['weight'])

        fsutil.write_file(weight_file, weight, fsync=False)
        logger.info('write: %s to file: %s' % (weight, weight_file))

    pids = conf.get('pids')

    if pids is None:
        return

    cgroup_util.add_pids(cgroup_path, pids)
    logger.info('add pids: %s to cgroup: %s' % (repr(pids), cgroup_path))


def reset_statistics(cgroup_path):
    reset_file = os.path.join(cgroup_path, 'blkio.reset_stats')
    fsutil.write_file(reset_file, '1', fsync=False)


def account(cgroup_path):
    file_name = os.path.join(cgroup_path, 'blkio.io_service_bytes_recursive')

    # file content format:
    # 8:0 Read 1053712384
    # 8:0 Write 81929383424
    # 8:0 Sync 80865775616
    # 8:0 Async 2117320192
    # 8:0 Total 82983095808
    # 8:16 Read 1053712384
    # 8:16 Write 81929383424
    # ...
    # Total 82983095808

    content = fsutil.read_file(file_name)
    lines = content.split('\n')[:-1]

    r = {}
    for line in lines:
        parts = line.split()
        if parts[1] in ('Read', 'Write'):
            r[parts[0]] = r.get(parts[0]) or {}
            r[parts[0]][parts[1]] = int(parts[2])

    return r




from kazoo.exceptions import BadVersionError
from kazoo.exceptions import NoNodeError

from pykit import utfjson
from pykit import ututil
from pykit import zktx
from pykit import zkutil
from pykit.zktx import ABORTED
from pykit.zktx import COMMITTED
from pykit.zktx import PURGED
from pykit.zktx.test import base

dd = ututil.dd


class TestZKStorage(base.ZKTestBase):

    def setUp(self):
        super(TestZKStorage, self).setUp()

        self.zk.create('tx', 'tx')
        self.zk.create('lock', '{}')
        self.zk.create('tx/txidset', '{}')
        self.zk.create('tx/journal', '{}')
        self.zk.create('record', '{}')

        zke, _ = zkutil.kazoo_client_ext(self.zk)
        self.zs = zktx.ZKStorage(zke)

    def test_txidset(self):

        self.zs.add_to_txidset(COMMITTED, 1)
        rst, ver = self.zs.txidset.get()

        self.assertEqual({COMMITTED: [[1, 2]],
                          ABORTED: [],
                          PURGED: [],
                          }, rst)

    def test_journal(self):

        self.zs.journal.create(1, {'a': 'b'})

        rst, ver = self.zs.journal.get(1)
        self.assertEqual({'a': 'b'}, rst)

        rst, ver = self.zk.get('tx/journal/0000000001')
        self.assertEqual({'a': 'b'}, utfjson.load(rst))

        self.assertRaises(BadVersionError, self.zs.journal.delete, 1, version=100)
        self.zs.journal.delete(1, version=0)

        self.assertRaises(NoNodeError, self.zs.journal.get, 1)

    def test_record(self):

        # nonode_callback
        rst, ver = self.zs.record.get('foo')
        self.assertEqual([(-1, None)], rst)

        self.zs.record.create('foo', [(1, 1)])

        # record_load
        rst, ver = self.zs.record.get('foo')
        self.assertEqual([[1, 1]], rst)

        self.zs.record.set('foo', [(1, 2)])
        rst, ver = self.zs.record.get('foo')
        self.assertEqual([[1, 2]], rst)

        # apply_record
        self.zs.apply_record(2, 'foo', 3)
        rst, ver = self.zs.record.get('foo')
        self.assertEqual([[1, 2], [2, 3]], rst)

    def test_lock(self):

        for holder, ver in self.zs.acquire_key_loop(2, 'foo', 10):
            self.fail("should not yield other lock holder")

        for holder, ver in self.zs.acquire_key_loop(2, 'foo', 10):
            self.fail("lock twice should be allowed.")

        try:
            for holder, ver in self.zs.acquire_key_loop(3, 'foo', 0.1):
                self.assertEqual(2, holder)
            self.fail('expect lock timeout')
        except zkutil.LockTimeout:
            pass

        # release

        released, txid, ver = self.zs.try_release_key(3, 'foo')
        self.assertEqual((False, 2), (released, txid))

        released, txid, ver = self.zs.try_release_key(2, 'foo')
        self.assertEqual((True, 2), (released, txid))

        # duplicated release
        released, txid, ver = self.zs.try_release_key(2, 'foo')
        self.assertEqual((True, 2), (released, txid))
